import random
import sys, os, re, uuid
import argparse
from datetime import datetime
import logging
import operator
import fnmatch
import subprocess
import shlex
import inspect, traceback, importlib
import pkg_resources
from collections import defaultdict, OrderedDict
from functools import reduce, wraps
from copy import deepcopy
import shelve
import numpy as np

from pymake.core import get_pymake_settings, reset_pymake_settings, PmkTemplate
from pymake.core.types import ExpDesign, ExpTensor, ExpSpace, Model, Corpus, Script, Spec, ExpVector
from pymake.core.types import ExpTensorV2 # @debug name integration
from pymake.core.format import ExpeFormat

from pymake.io import is_empty_file
from pymake.util.utils import colored, basestring, hash_objects



''' Grammar Expe '''



# Custom formatter
# From : https://stackoverflow.com/questions/14844970/modifying-logging-message-format-based-on-message-logging-level-in-python3
class MyLogFormatter(logging.Formatter):

    critical_fmt  = "====>>> CRITICAL: %(msg)s"
    err_fmt  = "===>> ERROR: %(msg)s"
    warn_fmt  = "==> Warning: %(msg)s"
    #info_fmt = '%(module): %(msg)s'
    #info_fmt = '%(levelno)d: %(msg)s'
    default_fmt = '%(msg)s'

    # Custom Log
    VDEBUG_LEVEL_NUM = 9
    logging.addLevelName(VDEBUG_LEVEL_NUM, "VDEBUG")
    logging.VDEBUG = 9

    def __init__(self, fmt=default_fmt):
        super().__init__(fmt=fmt, datefmt=None, style='%')

    def format(self, record):

        # Save the original format configured by the user
        # when the logger formatter was instantiated
        format_orig = self._style._fmt

        # Replace the original format with one customized by logging level
        if record.levelno == logging.WARNING:
            self._style._fmt = MyLogFormatter.warn_fmt
        elif record.levelno == logging.ERROR:
            self._style._fmt = MyLogFormatter.err_fmt
        elif record.levelno == logging.CRITICAL:
            self._style._fmt = MyLogFormatter.critical_fmt
        else:
            self._style._fmt = MyLogFormatter.default_fmt


        # Call the original formatter class to do the grunt work
        result = logging.Formatter.format(self, record)

        # Restore the original format configured by the user
        self._style._fmt = format_orig

        return result


def setup_logger(level=None, name='root'):
    #formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s')

    if level is None:
        level = 0

    if level == 1:
        level = logging.DEBUG
    elif level >= 2:
        level = logging.VDEBUG
    elif level == -1:
        level = logging.WARNING
    else:
        # make a --silent option for juste error and critial log ?
        level = logging.INFO

    # Get logger
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # Format handler
    handler = logging.StreamHandler(sys.stdout)
    #handler.setFormatter(logging.Formatter(fmt=fmt))
    handler.setFormatter(MyLogFormatter())

    # Set Logger
    logger.addHandler(handler)
    # Prevent logging propagation of handler,
    # who reults in logging things multiple times
    logger.propagate = False


    def vdebug(self, message, *args, **kws):
        # Yes, logger takes its '*args' as 'args'.
        if self.isEnabledFor(logging.VDEBUG):
            self._log(logging.VDEBUG, message, args, **kws)

    #logger.Logger.vdebug = vdebug
    logging.Logger.vdebug = vdebug

    return logger


class GramExp(object):
    ''' Create a mapping between different format of design of experiments.

        Attribute
        ---------
        _tensors :ExpTensor
            tensor of experiments

        Methods
        -------
        __init__ : conf choice in :
           * Expe -> One experiments
           * ExpTensor -> Design of experiments
           * Expe & spec in Expe -> Design of experiments with global value,
                either from conf or command-line args.
                conf value will udpdate all experiments first from command-line
                argument then from defaut values.

        Notes
        -----

        Design of Experiments can take three forms :
        1. command-line arguments (see self.parsearg).
            * use for parralization with Gnu parallel.
        2. Expe : Uniq experiments specificattion,
            * @Todo list of Expe
        3. ExpTensor : Represente a tensor of experiments.
           Each entry of the dict contains an iterable over specifications
           for this entry. (see frontend.io)
           * @Todo pass rule to filter experiments...

        ### Expe Filtering
        Expe can contains **special** keywords and value :
        * _bind rules : List of rules of constraintsi on the design.
            * Inclusif rules
                + [a.b]  --- a(value) shoudld occur only with b(value), (ie 'lda.fast_init')
                + [a.b.c] --- with a(value), key b(key) take only c (value), (ie 'lda.iterations.100')
            * Exclusif Rules
                + [a.!b]  --- a(value) shoudld not only with b(value),
                + [a.b.!c] --- with a(value), key b(key) don't take c (value),

            Warning : it does only check the last words when parameter are
                      separated by a dot (.) as for model module (name pmk.lda !) for example.

        * if an attribute  take the value "_null",
          then it will be removed from the Exp. This can be usefull when piping
          ouptuf of pymake to some script to parallelize.

        ### I/O Concerns
        An experiment typically write three kind of files :
        Corpus are load/saved using Pickle format in:
        * bdir/corpus_name.pk
        Models are load/saved using pickle/json/cvs in :
        * bdir/refdir/rep/model_name_parameters.pk   <--> ModelManager
        * bdir/refir/rep/model_name_parameters.json <--> DataBase
        * bdir/refdir/rep/model_name_parameters.inf  <--> ModelBase

    '''

    log = logging.getLogger('root')

    _version = pkg_resources.get_distribution('pmk').version

    _examples = '''Examples:
        >> Text fitting
        fit.py -k 6 -m ldafullbaye -p
        >> Network fitting :
        fit.py -m immsb -c alternate -n 100 -i 10'''

    _frontend_ext = ['gt', # graph-tool
                     'pk', # pickle
                    ]
    #_model_ext = @Todo: dense(numpy/pk.gz) or sparse => gt...?

    # has special semantics on **output_path**.
    _special_keywords = [ '_refdir', '_repeat',
                         '_format', '_csv_typo',
                        ] # output_path => pmk-basedire/{base_type}/{refdir}/{repeat}/${format}:${csv_typo}

    # Reserved by GramExp for expe identification
    _reserved_keywords = ['_spec', # for nested specification.
                          '_id_expe', # unique (locally) expe identifier.
                          '_name_expe', # exp name identifier, set before _default_expe
                          '_expe_hash', # Unique spec identifier.
                          '_pmk', # force to set some settings out of grammarg.
                         ]
    _private_keywords = _reserved_keywords + _special_keywords

    _default_expe = {
        #'_verbose'   : 0,
        '_write'     : False,

        '_ignore_format_unique' : False,

        '_force_load_data' : True, # if True, it will force the raw data parsing.
        '_force_save_data' : True, # if False, dont save corpus as pk/gt ont the filesystem.
        '_no_block_plot' : False,
    }


    _env = None # should be set during bootstrap
    _base_name = '.pmk'
    _cfg_name = 'pmk.cfg'
    _db_name = '.pmk-db'
    _pmk_error_file = '.pymake-errors'


    def __init__(self, conf, usage=None, parser=None, parseargs=True, expdesign=None):
        # @logger One logger by Expe ! # in preinit
        setup_logger(level=conf.get('_verbose'))

        if conf is None:
            conf = {}

        if parseargs is True:
            kwargs, self.argparser, _ = self.parseargsexpe(usage)
            conf.update(kwargs)
        if parser is not None:
            # merge parser an parsargs ?
            self.argparser = parser

        # It pollute the spec output...
        #[conf.update({k:v}) for k,v in self._default_expe.items() if k not in conf]

        self._base_conf = conf
        #self._do = conf.get('_do')

        self.exp_setup(conf, expdesign)

    @classmethod
    def getenv(cls, a=None):
        if a is None:
            return cls._env

        _env = cls._env
        return _env.get(a)

    @classmethod
    def setenv(cls, env):

        cls._env = env

        db = shelve.open(os.path.join(os.getcwd(), cls._db_name))
        db.update(env)
        db.close()

        cls._data_path = get_pymake_settings('project_data')

        if os.path.isfile(os.path.join(env.get('PWD'), cls._cfg_name)):
            cls._spec = Spec.get_all()
        else:
            cls._spec = {}


    def _update(self):
        # Seems obsolete, _conf is not writtent in _tensors, right ?
        self._conf = self._tensors._conf
        #
        self.lod = self._tensors._lod

    def update_default_expe(self, expformat):

        if not hasattr(expformat, '_default_expe'):
            return
        else:
            default_expe = expformat._default_expe

        # Use default _spec if no spec given
        _names = self.get_list('_name_expe')
        if len(_names) == 1 and '_default_expe' in _names and '_spec' in default_expe:
            names = default_expe['_spec']
            names = [names] if isinstance(names, str) else names
            group = []

            # Sensitive
            conf = self._base_conf
            #conf['_do'] = [self._do]

            for name in names:
                if isinstance(name, str):
                    d, expdesign = Spec.load(name, self._spec[name])
                    group.append((name, d, expdesign))
                else:
                    # Assume dict/expspace
                    group.append(('_expe', name, ExpDesign))
                    #raise NotImplementedError('use a defined spec in _default_expe')

            conf['_spec'] = group

            if 'script' in conf:
                script = conf.pop('script')
                _script, script_args = Script.get(script[0], script[1:])
                if script_args:
                    conf['_do'] = script_args
                    #self._tensors.update_all(_do=script_args)
                else:
                    conf['_do'] = ''
                    #self._tensors.update_all(_do=[''])

            self.exp_setup(conf)

        self._tensors.set_default_all(default_expe)

    def exp_setup(self, conf, expdesign=None):
        ''' work in self._tensors '''

        # Make main data structure
        self._tensors = ExpTensorV2.from_conf(conf, private_keywords=self._private_keywords,
                                              expdesign=expdesign)

        # Global settings (unique argument)
        self._conf = self._tensors.get_conf()

        # Set expTensor init.
        self._user_spec = conf.get('_spec', {})

        # makes it contextual.
        self._preprocess_exp()

        # Make lod
        skip_check = self._conf['_do'] in ['diff']
        self.lod = self._tensors.make_lod(skip_check)

        indexs = self._conf.get('_run_indexs')
        if indexs:
            self._tensors.remake(indexs)
            self._update()

    def _preprocess_exp(self):
        #self._check_exp(self._tensors)
        self._tensors.check_bind()
        self._tensors.check_model_typo()
        self._tensors.check_null()


    def io_format_check(self):
        if len(self) > 1 and '_write' in self._conf:
            self.check_format()

        # Clean pymake extra args:
        extra_args = ['_ignore_format_unique', ('_net', False)]
        keys_to_remove = []
        for _key in extra_args:
            if type(_key) is tuple:
                if self._conf.get(_key[0]) is _key[1]:
                    keys_to_remove.append(_key[0])
            elif _key in self._conf:
                    keys_to_remove.append(_key)
        # |
        for key in keys_to_remove:
            self._tensors.remove_all(key)


    def check_format(self):
        ''' Check if all expVector are distinguishable in _format.

            @debug: not valid check accros tensors !!!
            @debug: integration in ExpTensorV2 ?
        '''

        for tensor in self._tensors:
            hidden_key = []
            _format = tensor.get('_format', [''])
            if len(_format) > 1:
                raise NotImplementedError('multiple _format not implemented')
            else:
                _format = _format[0]

            format_settings = re.findall(r'{([^{}]*)}', _format)
            for setting, values in tensor.items():
                if setting in self._special_keywords:
                    continue
                if isinstance(values, list) and len(values) > 1 and setting not in format_settings:
                    hidden_key.append(setting)

            if _format and hidden_key and self._conf.get('_ignore_format_unique') is not True and not '_id' in format_settings:
                self.log.error('The following settings are not set in _format:')
                print(' '+ '  '.join(hidden_key))
                print('Possible conflicts in experience results outputs.')
                print('Please correct {_format} key to fit the experience settings.')
                print('To force the runs, use:  --ignore-format-unique')
                print('Exiting...')
                exit(2)

        if self._conf.get('_ignore_format_unique') is True:
            _hash = int((hash_objects(tensor)), 16) % 10**8
            _format = '{_name}-expe' + str(_hash) +'h'
            self._tensors.update_all(_format=_format)



    @classmethod
    def _check_exp(cls, tensor):
        ''' check format and rules of _tensors. '''
        for exp in tensor:
            # check reserved keyword are absent from exp
            for m in cls._reserved_keywords:
                if m in exp and m != '_name_expe':
                    raise ValueError('%s is a reserved keyword of gramExp.' % m)

            # Format
            assert(isinstance(exp, ExpTensor))
            for k, v in exp.items():
                if not issubclass(type(v), (list, tuple, set)):
                    raise ValueError('error, exp value should be iterable: %s' % k, v)

    def get_set(self, key, default=[]):
        ''' Return the (ordered) set of values of expVector of that {key}. '''

        # Manage special Values
        if key == '_spec':
            raise NotImplementedError
            ss =  sorted(self.get_nounique_keys())
            list_of_identifier = []
            for tensor in self._tensors:
                local_identifier = []
                for k in ss:
                    local_identifier.append(k)

                list_of_identifier.append('-'.join(sorted(filter(None,local_identifier))))
                _set = list_of_identifier

        else:
            _set = set()
            for v in self._tensors.get_all(key, default):
                if isinstance(v, (list, set, dict)):
                    self.log.debug('Unshashable value in tensor for key: %s' % key)
                    _set.add(str(v))
                else:
                    _set.add(v)

        return sorted(_set, key=lambda x: (x is None, x))


    def get_list(self, key, default=[]):
        ''' Return the list of values of expVector of that {key}. '''
        return self._tensors.get_all(key, default)

    def get_nounique_keys(self, *args):
        ''' return list of keys that are non unique in expe_tensor
            except if present in :args:.
        '''
        nk = self._tensors.get_nounique_keys()
        for o in args:
            if o in nk:
                nk.remove(o)

        return nk

    def get_array_loc(self, key1, key2, params, repeat=False):
        ''' Construct an 2d sink array.
            Return the zeros valued array of dimensions given by x and y {keys} dimension
        '''

        loc = []

        d1 = dict()
        for i, k in enumerate(self.get_set(key1)):
            d1[k] = i
        loc.append(d1)

        d2 = dict()
        for i, k in enumerate(self.get_set(key2)):
            d2[k] = i
        loc.append(d2)

        d3 = dict()
        for i, k in enumerate(params):
            d3[k] = i
        loc.append(d3)

        tr = lambda d,k:d[str(k)] if isinstance(k, (list, set, dict)) else d[k]

        floc = lambda k1, k2, z:(tr(d1,k1), tr(d2,k2), d3[z])
        shape = list(map(len, loc))
        if repeat:
            shape = [len(self.get_set('_repeat'))] + shape
        array = np.ma.array(np.empty(shape)*np.nan, mask=True)
        return array, floc


    def get_array_loc_n(self, keys, params):
        ''' Construct an nd sink array.
            Return the zeros valued array of dimensions given by x and y {keys} dimension
        '''

        loc = OrderedDict()

        for key in keys:

            d1 = dict()
            for i, v in enumerate(self.get_set(key)):
                d1[v] = i
            loc[key] = d1

        d2 = dict()
        for i, v in enumerate(params):
            d2[v] = i
        loc['_param_'] = d2

        def floc(expe, z):
            _special = ['_param_']
            pos = []
            for k in loc:
                if k in expe:
                    v = expe[k]
                elif k in _special:
                    continue
                else:
                    self.log.warning('indice unknown in floc.')
                    v = None

                pos.append(loc[k][v])

            pos.append(loc['_param_'][z])
            return tuple(pos)

        shape = [len(loc[d]) for d in loc]


        array = np.ma.array(np.empty(shape)*np.nan, mask=True)
        return array, floc

    def __len__(self):
        #return reduce(operator.mul, [len(v) for v in self._tensors.values()], 1)
        return len(self.lod)

    def make_forest_path(self, lod, ext, status='f', full_path=False):
        """ Make a list of path from a spec/dict, the filename are
            oredered need to be align with the get_from_conf_file.

            *args -> make_output_path
        """
        targets = []
        for spec in lod:
            filen = self.make_output_path(spec, ext, status=status, _null=self._tensors._null)
            if filen:
                s = filen.find(self._data_path)
                pt = 0
                if not full_path and s >= 0:
                    pt = s + len(self._data_path)
                targets.append(filen[pt:])
        return targets

    @classmethod
    def make_output_path(cls, expe, ext=None, status=None, _null=None, _nonunique=None):
        """ Make a single output path from a expe/dict
            @status: f finished
            @type: pk, json or inference.
        """
        expe = defaultdict(lambda: None, expe)

        dict_format = cls.transcript_expe(expe)
        if _null:
            dict_format.update(dict((k,None) for k in _null))

        basedir = os.path.join(cls._data_path, cls._base_name, 'results')

        hook = expe.get('_refdir', 'default')
        hook = hook.format(**dict_format)

        rep = ''
        if '_repeat' in expe and (expe['_repeat'] is not None and expe['_repeat'] is not False):
            rep = str(expe['_repeat'])
            if rep == '-1':
                rep = ''
            else:
                rep = rep.format(**dict_format)


        p = os.path.join(hook, rep)

        if not expe['_format']:
            # or give a hash if write is False ?
            cls.log.debug('No _format given, please set _format for output_path settings.')
            nonunique = ['_name_expe', '_do']
            if _nonunique:
                nonunique.extend(_nonunique)

            argnunique = []
            for i in set(nonunique):
                v = expe.get(i, 'oops')
                if isinstance(v, list):
                    v = v[0]
                argnunique.append(v)

            t = '-'.join(argnunique)
        else:
            t = expe['_format'].format(**dict_format)

        filen = os.path.join(basedir, p, t)

        if ext:
            filen = filen +'.'+ ext
        elif status and not ext:
            # Assume pickle driver for models.
            filen = filen +'.'+ 'pk'

        if status is 'f' and is_empty_file(filen):
            return  None
        else:
            return filen

    @classmethod
    def make_input_path(cls, expe, status=None):
        """ Make a single input path from a expe/dict """
        expe = defaultdict(lambda: None, expe)
        filen = None

        # Corpus is an historical exception and has its own subfolder.
        c = expe.get('corpus')
        if not c:
            c = ''
            cls.log.debug('No Corpus is given.')
        if c.lower().startswith(('clique', 'graph', 'generator')):
            c = c.replace('generator', 'Graph')
            c = c.replace('graph', 'Graph')
            c = 'generator/' + c
        if c.endswith(tuple('.'+ext for ext in cls._frontend_ext)):
            c = c[:-(1+len(c.split('.')[-1]))]

        input_dir = os.path.join(cls._data_path, cls._base_name, 'training', c)

        return input_dir

    @staticmethod
    def transcript_expe(expe):
        ''' Transcipt value to be used in spec's path formating '''

        fmt_expe = expe.copy()
        # iteration over {expe} trow a 'RuntimeError: dictionary changed size during iteration',
        # maybe due to **expe pass in argument ?

        id_str = 'expe' + str(expe['_id_expe'])
        id_name = expe['_name_expe']
        id_hash = expe['_expe_hash']

        # Special aliase for _format
        fmt_expe['_id'] = id_str
        fmt_expe['_name'] = id_name
        fmt_expe['_hash'] = id_hash

        for k, v in fmt_expe.items():
            if isinstance(v, (list, dict)):
                #fmt_expe[k] = id_str # don't do that, no robust get back expe.
                _hash = int((hash_objects(v)), 16) % 10**8
                fmt_expe[k] = k + str(_hash) +'h'
            elif isinstance(v, float):
                if 'e' in str(v):
                    fmt_expe[k] = str(v)
                elif len(str(v).split('.')[1]) > 2:
                    fmt_expe[k] = '%.2f' % v
                elif str(v).split('.')[1] == '0':
                    fmt_expe[k] = '%d' % v
                else:
                    fmt_expe[k] = v

        # Special rule to rewirte the output_path
        if '_model' in expe and 'model' in expe:
            fmt_expe['model'] = expe['_model']

        return fmt_expe



    @classmethod
    def get_parser(cls, description=None, usage=None):
        import pymake.core.gram as _gram
        parser = _gram.ExpArgumentParser(description=description, epilog=usage,
                                        formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument('-V', '--version', action='version', version='%(prog)s '+ cls._version)

        return parser

    @classmethod
    def push_gramarg(cls, parser, gram=None):
        import pymake.core.gram as _gram
        if gram is None:
            gram = _gram._Gram
        else:
            try:
                gram = importlib.import_module(gram)
            except ModuleNotFoundError as e:
                pr = get_pymake_settings('gramarg').split('.')[0]
                #@imporve priority shoiuld ne on : local project !
                cls.log.warning("Project name `%s' seems to already exists in your PYTHONPATH. Project's name should not conflict with existing ones." % pr)
                cls.log.warning("Please change the name of your project/repo")
                cls.log.warning('exiting.')
                exit(38)
            gram = next( (getattr(gram, _list) for _list in dir(gram) if isinstance(getattr(gram, _list), list)), None)

        grammar = []
        args_ = []
        for e in gram:
            if not isinstance(e, dict):
                args_.append(e)
                continue
            grammar.append((args_, e))
            args_ = []

        #[parser.add_argument(*r[0], **r[1]) for r in grammar]
        # check for duplicate
        for r in grammar:
            try:
                parser.add_argument(*r[0], **r[1])
            except argparse.ArgumentError as e:
                err_mesg = " (it'll be overwritten)"
                if r[0][0] in ('-v', '-s', '-w', '--seed'):
                    # @debug, level here is not set yet, so his not active by default.
                    # parsing difficulet becaus context sensitive (option (-v) impact how the options are interpreted)
                    cls.log.debug(str(e)+err_mesg)
                else:
                    cls.log.error(str(e)+err_mesg)
                #exit(3)


    @classmethod
    def parseargsexpe(cls, usage=None, args=None, parser=None):
        description = 'Launch and Specify Simulations.'
        if not usage:
            usage = GramExp._examples

        if parser is None:
            parser = cls.get_parser(description, usage)
        #s, remaining = parser.parse_known_args(args=args)

        # Push pymake grmarg.
        cls.push_gramarg(parser)
        # third-party
        gramarg = get_pymake_settings('gramarg')
        if gramarg:
            cls.push_gramarg(parser, gramarg)

        s = parser.parse_args(args=args)

        # Assume None value are non-filled options
        settings = dict((key,value) for key, value in vars(s).items() if value is not None)

        # Purge default/unchanged settings
        for k in list(settings):
            v = settings[k]
            if cls._default_expe.get(k) == v:
                settings.pop(k)

        pmk_opts = cls.pmk_extra_opts(settings)
        settings.update(pmk_opts)
        expdesign = GramExp.expVectorLookup(settings)

        return settings, parser, expdesign

    @staticmethod
    def pmk_extra_opts(settings):
        opts = {}
        pmk = settings.get('_pmk')
        if not pmk:
            return opts

        for s in pmk:
            k, v = s.split('=')
            try:
                v = int(v)
            except:
                try: v = float(v)
                except: pass
            opts[k] = v

        return opts

    @classmethod
    def zymake(cls, request={}, usage='', firsttime=True, expdesign=None):
        usage ='''\
        ----------------
        Communicate with the data :
        ----------------
         |   pmk update                                       : update the pymake index
         |   pmk -l [spec(default)|model|script|topo]         : list available component
         |   pmk show SPEC                                    : show one spec details (default if no arguments)
         |   pmk run SPEC [--script [fun][*args]] ...         : execute tasks (default if -x is given)
         |   pmk runpara SPEC [--script [fun][*args]] ...     : parallelize tasks (implicit if --cores is given)
         |   pmk hist [-n n_lines]                            : show command history
         |   pmk cmd SPEC ...                                 : Try to generate the command-line for each expe.
         |   pmk path SPEC Filetype(pk|json|inf) [status] ... : show output_path of each expe.
        ''' + '\n' + usage

        s, parser, expdesign_lkp = GramExp.parseargsexpe(usage)
        request.update(s)

        ontology = dict(
            _do    = ['cmd', 'show', 'path', 'run', 'update', 'init', 'runpara', 'hist', 'diff'],
            _spec   = list(cls._spec),
            _ext = ['json', 'pk', 'inf']
        )
        ont_values = sum([w for k, w in ontology.items() if k != '_spec'] , [])

        # Init _do value.
        if not request.get('_do'):
            request['_do'] = []

        # Special Case for CLI.
        if 'script' in request:
            # check if no command is specified, and
            # if 'script" is there, set 'run' command as default.
            do = request.get('_do', [])
            no_do = len(do) == 0
            no_do_command = len(request['_do']) > 0 and not request['_do'][0] in ontology['_do']
            if no_do or no_do_command:
                if '_cores' in request or '_net' in request:
                    runcmd = 'runpara'
                else:
                    runcmd = 'run'
                request['_do'] = [runcmd] + do


        do = request.get('_do', [])
        checksum = len(do)
        # No more Complex !
        run_indexs = []
        expgroup = []
        for i, v in enumerate(do.copy()):
            if str.isdigit(v):
                run_indexs.append(int(v))
                do.remove(v)
                checksum -= 1
            else:
                for ont, words in ontology.items():
                    # Walktrough the ontology to find arg semantics
                    if v in words:
                        if ont == '_spec':
                            if v in ont_values:
                                cls.log.critical('=> Warning: conflict between name of ExpDesign and Pymake commands')
                            do.remove(v)
                            d, expdesign = Spec.load(v, cls._spec[v])
                            expgroup.append((v,d, expdesign))
                        else:
                            request[ont] = v

                        checksum -= 1
                        break

        # Check erros in the command line
        if checksum != 0:
            if (request.get('_do') or request.get('do_list')) and not GramExp.is_pymake_dir():
                print('fatal: Not a pymake directory: %s not found.' % (cls._cfg_name))
                try:
                    os.remove('.pmk-db.db')
                except FileNotFoundError:
                    pass
                exit(10)

            if  firsttime == True:
                cls.log.warning('Spec not found, re-building Spec indexes...')
                cls.update_index('spec')
                cls._spec = Spec.get_all()
                return cls.zymake(firsttime=False)
            else:
                cls.log.error('==> Error : unknown argument: %s\n\nAvailable Exp : %s' % (do, list(cls._spec)))
                exit(10)


        # Setup the exp inputs
        if run_indexs:
            request['_run_indexs'] = run_indexs

        if len(expgroup) >= 1:
            request['_spec'] = expgroup

        if len(expgroup) > 0 and len(do) == 0:
            request['_do'] = 'show'


        expdesign = expdesign or expdesign_lkp
        return cls(request, usage=usage, parser=parser, parseargs=False, expdesign=expdesign)


    @classmethod
    def expVectorLookup(cls, request):
        ''' set exp from spec if presents '''

        expdesign = None
        for k, v in request.items():
            if not isinstance(v, ExpVector):
                continue

            sub_request = ExpVector()
            for vv in v:
                if isinstance(vv, list):
                    continue
                if vv in cls._spec:
                    loaded_v, expdesign = Spec.load(vv, cls._spec[vv])
                    if isinstance(loaded_v, ExpVector):
                        sub_request.extend(loaded_v)
                else:
                    # Multiple Flags
                    sub_request.append(vv)

            if sub_request:
                request[k] = sub_request

        return expdesign


    @classmethod
    def exp_tabulate(cls, conf={}, usage=''):

        gargs = clargs.grouped['_'].all
        for arg in gargs:
            try:
                conf['K'] = int(arg)
            except:
                conf['model'] = arg
        return cls(conf, usage=usage)

    def make_path(self, ext=None, status=None, fullpath=None):
        return self.make_forest_path(self.lod, ext, status, fullpath)

    #@deprecated
    def make_commandline(self):
        commands = self.make_commands()
        commands = [' '.join(c) for c in commands]
        return commands

    #@deprecated
    def make_commands(self):
        lod = self.lod
        commands = []
        for e in lod:
            argback = self.argparser._actions
            command = []
            args_seen = []
            for a in argback:
                if not hasattr(a, 'option_strings') or len(a.option_strings) == 0:
                    # Assume every options of expe has a flags
                    continue
                if a.dest in e and a.dest not in args_seen:
                    #if isinstance(a, (gram.unaggregate_append,)):
                    #    # Assume except_append are removed from expe,
                    #    # special traitment
                    #    continue
                    if a.nargs == 0:
                        # standalone flags
                        store = e[a.dest]
                        if 'StoreTrue' in str(type(a)):
                            storetrue = True
                        elif 'StoreFalse' in str(type(a)):
                            storetrue = False
                        else:
                            raise ValueError('Check the Type of argparse for : %s'%e)
                        if storetrue and store:
                            command += [a.option_strings[0]]
                    else:
                        _arg = e[a.dest]
                        if isinstance(_arg, (list, set, tuple)):
                            _arg = ' '.join(str(_arg))
                        command += [a.option_strings[0]] + [str(_arg)]
                        args_seen.append(a.dest)

            commands.append(command)
        return commands

    def reorder_firstnonvalid(self, ext='pk'):
        for i, e in enumerate(self.lod):
            if not self.make_output_path(e, ext, status='f', _null=self._tensors._null, _nonunique=self.get_nounique_keys()):
                self.lod[0], self.lod[i] = self.lod[i], self.lod[0]
                break
        return

    def exptable(self):
        return self._tensors.table()

    def spectable(self):
        return Spec.table()

    def scripttable(self):
        return Script.table()

    def modeltable(self, _type='short'):
        return Model.table(_type)

    def spectable_topo(self):
        return Spec.table_topos(self._spec)

    def alltable_topo(self):
        from pymake.core.types import _table_
        specs = self._spec
        scripts = Script.get_all()
        models = Model.get_all()
        table = [models, specs, scripts]
        headers = ['Models', 'Specs', 'Actions']
        return _table_(table, headers)

    def help_short(self):
        shelp = self.argparser.format_usage() + self.argparser.epilog
        return shelp

    def simulate_short(self):
        ''' Simulation Output '''
        print('''
              Nb of experiments : %s
              Corpuses : %s
              Models : %s
              ''' % (len(self), self.getCorpuses(), self.getModels()), file=sys.stdout)
        exit()

    def simulate(self, halt=True, file=sys.stdout):

        print('PYMAKE Exp: %d' % (len(self) ), file=file)
        print('-'*30, file=file)
        print(self.exptable(), file=file)
        if halt:
            exit()
        else:
            return

    @staticmethod
    def sign_nargs(fun):
        return sum([y.default is inspect._empty for x,y in inspect.signature(fun).parameters.items() if x != 'self'])

    @staticmethod
    def tb_expeformat(sandbox):
        signatures = []
        for m in dir(sandbox):
            if not callable(getattr(sandbox, m)) or m.startswith('__') or hasattr(ExpeFormat, m):
                continue
            sgn = inspect.signature(getattr(sandbox, m))
            d = [v for k,v in sgn.parameters.items() if k != 'self'] or []
            signatures.append((m, d))
        return signatures

    @staticmethod
    def functable(obj):
        ''' show method/doc associated to one class (in /scripts) '''
        lines = []
        for m in  GramExp.tb_expeformat(obj):
            name = m[0]
            opts = []
            for o in m[1]:
                if not o:
                    continue
                if o.default is inspect._empty:
                    # args
                    opts.append(o.name)
                else:
                    # kwargs
                    opts.append('%s [%s]'% (o.name, o.default))
            opts = ', '.join(opts)
            line = '  ' + name + ': ' + opts
            lines.append(line)
        return lines

    @staticmethod
    def load(*args, **kwargs):
        from pymake.io import load
        return load(*args, **kwargs)

    @staticmethod
    def save(*args, **kwargs):
        from pymake.io import save
        return save(*args, **kwargs)


    @staticmethod
    def model_walker(bdir, fmt='list'):
        models_files = []
        if fmt == 'list':
            ### Easy formating
            for root, dirnames, filenames in os.walk(bdir):
                for filename in fnmatch.filter(filenames, '*.pk'):
                    models_files.append(os.path.join(root, filename))
            return models_files
        else:
            ### More Complex formating
            tree = { 'json': [],
                    'pk': [],
                    'inference': [] }
            for filename in fnmatch.filter(filenames, '*.pk'):
                if filename.startswith(('dico.','vocab.')):
                    dico_files.append(os.path.join(root, filename))
                else:
                    corpus_files.append(os.path.join(root, filename))
            raise NotImplementedError()
        return tree


    @staticmethod
    def get_cls_name(cls):
        clss = str(cls).split()[1]
        return clss.split('.')[-1].replace("'>", '')


    def expe_init(self, expe, _seed_path='/tmp/pmk.seed'):
        ''' Intialize an expe:
            * Set seed
            * set in/out filesystem path
        '''

        _seed = expe.get('_seed')

        if _seed is None:
            seed0 = random.randint(0, 2**128)
            seed1 = np.random.randint(0, 2**32, 10)
            seed = [seed0, seed1]
        elif type(_seed) is str and str.isdigit(_seed):
            _seed = int(_seed)
            seed = [_seed, _seed]
        elif type(_seed) is str:
            if _seed in expe:
                _seed = expe[_seed]

            seed = []
            for c in list(_seed):
                seed.append(str(ord(c)))

            if '_repeat' in expe:
                if type(expe['_repeat']) is int:
                    seed.append(str(expe['_repeat']))
                else:
                    for c in list(expe['_repeat']):
                        seed.append(str(ord(c)))

            seed = ''.join( [chr(int(i)) for i in list(''.join(seed))] )
            seed = int((hash_objects(seed)), 32) % 2**32
            seed = [seed, seed]
            # Makes it on 32 bit...

        elif _seed is True:
            # Load state
            seed = None
            try:
                self._seed = self.load(_seed_path)
                random.seed(self._seed[0])
                np.random.seed(self._seed[1])
                #np.random.set_state(seed_state)
            except FileNotFoundError as e:
                self.log.error("Cannot initialize seed, %s file does not exist." % _seed_path)
                sid = [np.random.randint(0, 2**63), np.random.randint(0, 2**32)]
                self.save(_seed_path, sid, silent=True)
                #self.save(_seed_path, np.random.get_state() silent=True)
                raise FileNotFoundError('%s file just created, try again !')

        if seed:
            # if no seed is given, it's impossible to get a seed from numpy
            # https://stackoverflow.com/questions/32172054/how-can-i-retrieve-the-current-seed-of-numpys-random-number-generator
            random.seed(seed[0])
            np.random.seed(seed[1])

            # Save state
            self.save(_seed_path, seed, silent=True)
            #self.save(_seed_path, [seed, np.random.get_state()], silent=True)
            self._seed = seed

        # Init I/O settings
        expe['_output_path'] = self.make_output_path(expe, _null=self._tensors._null, _nonunique=self.get_nounique_keys())
        expe['_input_path'] = self.make_input_path(expe)

        return self._seed


    def execute(self):
        ''' Execute Exp Sequentially '''

        if 'script' in self._conf:
            script = self._conf.pop('script')
            self._tensors.remove_all('script')
            self._tensors.remove_all('_do')
        else:
            self.log.error('==> Error : You need to specify a script. (--script)')
            exit(10)

        try:
            _script, script_args = Script.get(script[0], script[1:])
        except (ValueError, TypeError) as e:
            self.log.warning('Script not found, re-building Scripts indexes...')
            self.update_index('script')
            #self._spec = Spec.get_all()
            try:
                res = Script.get(script[0], script[1:])
                if not res:
                    print('error: Unknown script: %s' % (script[0]))
                    exit(404)
                else:
                    _script, script_args = res
            except:
                raise
        except IndexError as e:
                self.log.error('Script arguments error : %s -- %s' % (e, script))
                exit(2)

        # Raw search
        #script_name = script[0]
        #script_args = script[1:]
        #Scripts = mloader.ScriptsLoader.get_packages()
        #if not script_name in Scripts:
        #    method_by_cls = mloader.ScriptsLoader.lookup_methods()
        #    if script_name in sum(method_by_cls.values(), []):
        #        # check if it is a method reference
        #        script_args = [script_name] + script_args
        #        script_name = next(k.lower() for (k,v) in method_by_cls.items() if script_name in v)
        #    else:
        #        raise ValueError('error: Unknown script: %s' % (script_name))

        if script_args:
            self._tensors.update_all(_do=script_args)
        #self.pymake(sandbox=Scripts[script_name])
        self.pymake(sandbox=_script)

    def execute_parallel(self):

        basecmd = sys.argv.copy()
        #try:
        #    Target = subprocess.check_output(['which','pymake']).strip().decode()
        #except:
        #    Target = 'python3 /home/ama/adulac/.local/bin/pymake'
        Target = 'pmk'
        basecmd = [Target] + basecmd[1:]

        # Create commands indexs
        indexs = self._conf.get('_run_indexs')
        if not indexs:
            indexs = range(len(self))
        else:
            [basecmd.remove(str(i)) for i in indexs]

        # Creates a list of commands that pick each index
        # from base requests commands.
        cmdlines = []
        for index in indexs:
            id_cmd = 'run %s' % (index)
            if 'runpara' in basecmd:
                cmd = ' '.join(basecmd).replace('runpara', id_cmd, 1)
            else:
                idx = basecmd.index(Target)
                cmd = ' '.join(basecmd[:idx] + [Target + ' '+ id_cmd] + basecmd[idx+1:])
            cmdlines.append(cmd)

        n_cores = str(self._conf.get('_cores', 1))

        # remove the --cores options
        for i, cmd in enumerate(cmdlines):
            cmd = cmd.split()
            try: idx = cmd.index('--cores')
            except: continue
            cmd.pop(idx); cmd.pop(idx) # pop --cores int
            cmdlines[i] = ' '.join(cmd)


        if self._conf.get('simulate'):
            self.simulate()

        #for r in cmdlines:
        #    print(r)
        #exit()

        cmd = ['parallel', '-j', n_cores, '-u', '-C', "' '", '--eta', '--progress', ':::', '%s'%('\n'.join(cmdlines))]

        #stdout = subprocess.check_output(cmd)
        #print(stdout.decode())
        for line in self.subexecute1(cmd):
            print(line, end='')

        #self.subexecute2(cmd)

    def execute_parallel_net(self, nhosts=None):
        ''' run X process by machine !
            if :nhosts: (int) is given, limit the numner of remote machines.
        '''

        NDL = get_pymake_settings('loginfile')
        workdir = get_pymake_settings('remote_pwd')
        remotes = list(filter(None, [s for s in open(NDL).read().split('\n') if not s.startswith(('#', '%'))]))

        basecmd = sys.argv.copy()
        #Target = './zymake.py'
        #basecmd = ['python3', Target] + basecmd[1:]
        Target = 'pmk'
        basecmd = [Target] + basecmd[1:]
        cmdlines = None

        # Create commands indexs
        indexs = self._conf.get('_run_indexs')
        if not indexs:
            indexs = range(len(self))
        else:
            [basecmd.remove(str(i)) for i in indexs]

        # Get chunked indexs
        if nhosts is None:
            # cores by host
            n_cores = int(self._conf.get('_cores', 1))
        else:
            # share run per host
            n_cores = len(indexs) // int(nhosts)

            # Only --net 1, implement/understood from now.
            if int(nhosts) == 1:
                indexs = []
                cmdlines = [' '.join(basecmd)]
            else:
                raise NotImplementedError

        if cmdlines is None:
            indexs = list(map(str, indexs))
            chunk = int(np.ceil(len(indexs) / len(remotes)))
            indexs = [indexs[i:i+chunk] for i in range(0, len(indexs), chunk)]

            # Creates a list of commands that pick each index
            # from base requests commands.
            cmdlines = []
            for index in indexs:
                id_cmd = '%s' % (' '.join(index))
                if 'runpara' in basecmd:
                    cmd = ' '.join(basecmd).replace('runpara', id_cmd, 1)
                else:
                    idx = basecmd.index(Target)
                    cmd = ' '.join(basecmd[:idx] + [Target + ' '+ id_cmd] + basecmd[idx+1:])

                cmdlines.append(cmd)

        # remove the --net options
        for i, cmd in enumerate(cmdlines):
            #cmdlines[i] = cmd.replace('--net', '').strip()
            cmdlines[i] = re.sub(r'--net\s*[0-9]*', '', cmd).strip()

        if nhosts is not None:
            tempf = '/tmp/pmk_' + uuid.uuid4().hex
            nhosts = int(nhosts)
            with open(tempf, 'w') as _f_w:
                with open(NDL) as _f_r:
                    for i, l in enumerate(_f_r.readlines()):
                        if i >= nhosts:
                            break
                        _f_w.write(l)
            NDL = tempf

        #for r in cmdlines:
        #    print(r)
        #exit()

        cmd = ['parallel', '-u', '-C', "' '", '--eta', '--progress',
               '--sshloginfile', NDL, '--workdir', workdir,
               '--env', 'OMP_NUM_THREADS', '--env', 'PYTHONPATH', '--env', 'PATH',
               ':::', '%s'%('\n'.join(cmdlines))]

        env = {'PYTHONPATH': '~/.local/lib/:',
               'PATH': '~/.local/bin:/usr/local/bin:/usr/bin:/bin:'}

        if self._conf.get('simulate'):
            self.simulate()


        #stdout = subprocess.check_output(cmd)
        #print(stdout.decode())
        for line in self.subexecute1(cmd, **env):
            print(line, end='')

    @staticmethod
    def subexecute1(cmd, **env):
        _env = os.environ
        _env.update(env)
        popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True, env=_env)
        for stdout_line in iter(popen.stdout.readline, ""):
            yield stdout_line
        popen.stdout.close()
        return_code = popen.wait()
        if return_code:
            raise subprocess.CalledProcessError(return_code, cmd)

    @staticmethod
    def subexecute2(cmd):
        ''' trying to print colors here ...'''
        popen = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, universal_newlines=True)
        while True:
            output = popen.stdout.readline()
            if output == '' and popen.poll() is not None:
                break
            if output:
                print(popen.strip())
        return process.poll()

    def notebook(self):
        from nbformat import v4 as nbf
        nb = nbf.new_notebook()

        text = ''
        # get the expe
        # get the script
        # get the figure
        code = ''
        nb['cells'] = [nbf.new_markdown_cell(text),
                       nbf.new_code_cell(code) ]
        return

    @classmethod
    def is_pymake_dir(cls):
        pwd = cls.getenv('PWD')
        return os.path.isfile(os.path.join(pwd, cls._cfg_name))

    def init_folders(self):
        join = os.path.join

        if self.is_pymake_dir():
            print('fatal: %s file already exists.' % (self._cfg_name))
            exit(11)

        pwd = self.getenv('PWD')
        cwd = os.path.dirname(__file__)
        folders = ['_current', 'spec', 'script', 'model']
        #open(join(pwd, '__init__.py'), 'a').close()
        spec = {'projectname':os.path.basename(pwd)}
        print('Creating project: {projectname}'.format(**spec))

        settings = {}
        for d in folders:
            if d in ['model', 'spec', 'script']:
                os.makedirs(d, exist_ok=True)
                with open(join(cwd,'..', 'template', '%s.template'%(d))) as _f:
                    template = PmkTemplate(_f.read())
                open(join(pwd, d,  '__init__.py'), 'a').close()
                with open(join(pwd, d,  'template_%s.py'%(d)), 'a') as _f:
                    _f.write(template.substitute(spec))
                settings.update({'default_%s'%(d):'.'.join((spec['projectname'], d))})
            elif d == '_current':
                # Gramarg
                with open(join(cwd,'..', 'template', 'gramarg.template')) as _f:
                    template = PmkTemplate(_f.read())
                with open(join(pwd, 'gramarg.py'), 'a') as _f:
                    _f.write(template.substitute(spec))
                settings.update({'default_gramarg':'.'.join((spec['projectname'], 'gramarg'))})
            else:
                raise ValueError('Doh, Directory unknwow: %s' % d)


        # Set and write pymake.cfg
        reset_pymake_settings(settings)

        return self.update_index()

    def pushcmd2hist(self):
        from pymake.util.utils import tail
        bdir = os.path.join(self._data_path, self._base_name)
        fn = os.path.join(bdir, '.pymake-hist')
        if not os.path.isfile(fn):
            open(fn, 'a').close()
            return

        cmd = sys.argv.copy()
        cmd[0] = os.path.basename(cmd[0])
        cmd = ' '.join(cmd)
        _tail = tail(fn, 10)

        with open(fn, 'a') as _f:
            if not cmd in _tail:
                return _f.write(cmd+'\n')


    def show_history(self):
        from pymake.util.utils import tail
        n = self._conf.get('N', 42)
        if n == 'all':
            n_lines = -1
        else:
            n_lines = int(n)

        bdir = self._data_path
        fn = os.path.join(bdir, self._base_name, '.pymake-hist')
        if not os.path.isfile(fn):
            self.log.error('hist file does not exist.')
            return

        _tail = tail(fn, n_lines)

        print('\n'.join(_tail))

    def show_diff(self):
        from tabulate import tabulate
        import itertools

        diff_expe = dict()
        for tensor in self._tensors:
            for k in self._tensors.get_keys():

                if k not in diff_expe:
                    diff_expe[k] = tensor.get(k, ['--'])
                    continue

                v = tensor.get(k, ['--'])
                s = diff_expe[k]
                r = list(itertools.filterfalse(lambda x: x in v, s)) + list(itertools.filterfalse(lambda x: x in s, v))
                #try:
                #    r = (set(s) - set(v)).union(set(v) - set(s))
                #except TypeError as e:
                #    r = list(itertools.filterfalse(lambda x: x in v, s)) + list(itertools.filterfalse(lambda x: x in s, v))


                diff_expe[k] = r


        for k in list(diff_expe):
            if len(diff_expe[k]) == 0:
                diff_expe.pop(k)

        if diff_expe:
            print('exp differs:')
            print(tabulate(diff_expe.items()))

        exit()


    @classmethod
    def update_index(cls, *index_name):
        from pymake.index.indexmanager import IndexManager as IX

        pwd = cls.getenv('PWD')
        os.chdir(pwd)

        ## Update index
        if len(index_name) == 0:
            IX.build_indexes()
        else:
            for name in index_name:
                IX.build_indexes(name)

        ## Update bash_completion file
        home = os.path.expanduser('~')
        cwd = os.path.dirname(__file__)
        pjt = os.path.basename(pwd)

        template = None
        completion_fn = os.path.join(home, '.bash_completion.d', 'pymake_completion')
        if os.path.exists(completion_fn):
            with open(completion_fn) as _f:
                template = _f.read()

            # Reset completion file if version differs
            verpos = template.find('%%PMK')
            _ver = re.search(r'version=([0-9\.\-a-zA-Z_]*)', template[verpos:])
            if not _ver:
                template = None
            else:
                _ver = _ver.groups()[0]
                if _ver != cls._version:
                    template = None

        if template is None:
            with open(os.path.join(cwd, '..', 'template', 'pymake_completion.template')) as _f:
                template = _f.read()


        # Get Specs
        specs = ' '.join(list(cls._spec))

        # Get Scripts
        _scripts = Script.get_all(_type='hierarchical')
        scripts = defaultdict(list)
        all_scripts = set()
        sur_scripts = set()
        for _o in _scripts:
            script = _o['scriptsurname']
            action = _o['method']
            scripts[script].append(action)
            all_scripts.update([script, action])
            sur_scripts.add(script)

        # Create a Bash array of strings.
        dict_scripts = []
        for sur in sur_scripts:
            dict_scripts.append('"'+ ' '.join(scripts[sur]) +'"')

        # get Models
        models = None

        # get Corpus
        corpus = None


        all_scripts = ' '.join(all_scripts)
        sur_scripts = ' '.join(sur_scripts)
        dict_scripts = '(' + ' '.join(dict_scripts) + ')'
        hook = '''
                elif [[ "$project" == "$$projectname" ]]; then
                    specs="$$specs"
                    all_scripts="$$all_scripts"
                    sur_scripts="$$sur_scripts"
                    dict_scripts=$$dict_scripts
               '''

        _match = '[[ "$project" == "%s" ]]' % (pjt)
        back_pos = template.find(_match)
        if back_pos  >= 0:
            # Remove old lines
            template = template.split('\n')
            pt = None
            for pos, line in enumerate(template):
                if _match in line:
                    pt = pos
                    break
            _hook = hook.strip().split('\n')
            n_lines = len(_hook)
            template = template[:pt] +_hook+ template[pt+n_lines:]
            template = '\n'.join(template)
        else:
            insert_pos = template.find('%%PMK')
            insert_pos = insert_pos - template[:insert_pos][::-1].find('\n')
            template = template[:insert_pos] +\
                    hook +\
                    template[insert_pos:]

        os.makedirs(os.path.join(home, '.bash_completion.d'), exist_ok=True)
        with open(completion_fn, 'w') as _f:
            template = PmkTemplate(template)
            template = template.substitute(projectname=pjt, version=cls._version,
                                           specs=specs,
                                           all_scripts=all_scripts,
                                           sur_scripts=sur_scripts,
                                           dict_scripts=dict_scripts)
            _f.write(template)

        #os.execl("/bin/bash", "/bin/bash", "-c", "source ~/.bash_completion.d/pymake_completion")

    def pymake(self, sandbox=ExpeFormat):
        ''' Walk Through experiments. '''

        if 'do_list' in self._conf:
            print('Available methods for %s: ' % (sandbox))
            print(*self.functable(sandbox) , sep='\n')
            exit()

        # Default spec, expVector ?
        #if hasattr(sandbox, '_expe_default'):
        #    print(sandbox._expe_default)

        sandbox._preprocess_(self)
        self.io_format_check()

        if self._conf.get('simulate'):
            self.simulate()

        n_errors = 0
        for id_expe, _expe in enumerate(self.lod):
            expe = ExpSpace(**_expe)

            pt = dict((key, value.index(expe[key])) for key, value in self._tensors.get_gt().items()
                      if (isinstance(expe.get(key), (basestring, int, float)) and key not in self._reserved_keywords))
            pt['expe'] = id_expe

            # Init Expe
            expdesign = self._tensors._ds[id_expe]
            self.expe_init(expe)

            try:
                expbox = sandbox(pt, expe, expdesign, self)
            except FileNotFoundError as e:
                self.log.error('ignoring %s'%e)
                continue
            except Exception as e:
                print(('Error during '+colored('%s', 'red')+' Initialization.') % (str(sandbox)))
                traceback.print_exc()
                exit(2)

            # Expe Preprocess
            expbox._expe_preprocess()

            # Setup handler
            if '_do' in expe and len(expe._do) > 0:
                # ExpFormat task ? decorator and autoamtic argument passing ....
                do = expe._do
                pmk = getattr(expbox, do[0])
            else:
                do = ['__call__']
                pmk = expbox

            # Launch handler
            args = do[1:]
            try:
                res = pmk(*args)
                if res is not None:
                    print(res)
            except KeyboardInterrupt:
                # it's hard to detach matplotlib...
                traceback.print_exc(file=sys.stdout)
                break
            except Exception as e:
                n_errors += 1
                self.log.critical(('Error during '+colored('%s', 'red')+' Expe no %d.') % (do, id_expe))
                traceback.print_exc()
                with open(self._pmk_error_file, 'a') as _f:
                    lines = []
                    lines.append('%s' % (datetime.now()))
                    lines.append('Error during %s Expe no %d' % (do, id_expe))
                    lines.append('Output path: %s' % (expe.get('_output_path')))
                    _f.write('\n'.join(lines) + '\n')
                    traceback.print_exc(file=_f)
                    _f.write('\n')
                continue

            # Expe Postprocess
            expbox._expe_postprocess()

        if n_errors > 0:
            self.log.warning("There was %d errors, logged in `%s'" % (n_errors, self._pmk_error_file))
            with open(self._pmk_error_file, 'a') as _f:
                _f.write(100*'='+'\n')

        sandbox._postprocess_(self)
        self.pushcmd2hist()

        exit_status = 0 if n_errors == 0 else 1
        return exit(exit_status)







#!/usr/bin/env python
import os
import sys

import numpy as np
import pickle
from scipy.misc import logsumexp
try:
    from sympy.functions.combinatorial.numbers import stirling
    import sympy as sym
except:
    pass

from pymake.util.utils import get_pymake_settings


def load_stirling(style='npy'):
    stirling_path = get_pymake_settings('project_stirling')
    fn = os.path.join(stirling_path,'stirling.npy')
    npy_exists = os.path.isfile(fn)
    if style == 'npy' and npy_exists:
        return np.load(fn)
    else:
        stirlg = lookup_stirling()
        return stirlg.load()

class lookup_stirling(object):
    def __init__(self, k=1000, fn='stirling.pk' ):
        sys.setrecursionlimit(100000)
        self.fn = os.path.join(os.path.dirname(__file__),fn)
        self.k_max = k
        self.kind = 1

    def _reset(self, k):
        self.k_max = k
        self.array_stir = np.ones((k, k)) * np.inf
        self.array_stir[0,0] = 0

    def _stirling_table_dishe(self, n, m):
        if m > n:
            return np.inf
        else:
            return sym.log(stirling(n, m, kind=self.kind)).evalf()

    def run(self, k=None, save=True):
        _f = open(self.fn, 'wb')
        k_max = k or self.k_max
        self._reset(k_max)

        # Run stirling computation for (n,m) matrix equal to stirling(n,m) if n <= m else 0
        for n in range(k_max):
            print( n)
            self.array_stir[n, 1:] = np.array([ self._stirling_table_dishe(n, m) for m in range(1, k_max) ])
            if save:
                pickle.dump(self.array_stir[n], _f, protocol=pickle.HIGHEST_PROTOCOL)
                _f.flush()

        _f.close()
        return self.array_stir

    def load(self, fn=None):
        fn = fn or self.fn
        array_stir = np.array([])
        with open(fn, "rb") as f:
            array_stir = np.hstack((array_stir, pickle.load(f)))
            while True:
                try:
                    # Don't get why call to load is needed twice, but get 0.0 otherwise
                    array_stir = np.vstack((array_stir, pickle.load(f)))
                except EOFError: break
                except ValueError: break

        self.k_max = len(array_stir)
        self.array_stir = array_stir
        return self.array_stir

    def recursive_line(self, new_line=5246):
        stir = self.load()
        J = stir.shape[0]
        K = stir.shape[1]
        for x in range(new_line):
            n = J + x
            new_l =  np.ones((1, K)) * np.inf
            print(n)
            for m in range(1,K):
                if m > n:
                    continue
                elif m == n:
                    new_l[0, m] = 0
                elif m == 1:
                    new_l[0, 1] = logsumexp( [  np.log(n-1) + stir[n-1, m] ] )
                else:
                    new_l[0, m] = logsumexp( [ stir[n-1, m-1] , np.log(n-1) + stir[n-1, m] ] )
            stir = np.vstack((stir, new_l))

        #np.save('stirling.npy', stir)
        #np.load('stirling.npy')
        return stir

    def recursive_row(self, new_row=''):
        stir = np.load('stirling.npy')
        J = stir.shape[0]
        K = stir.shape[1]
        x = 0
        while stir.shape[0] != stir.shape[1]:
            m = K + x
            new_c =  np.ones((J, 1)) * np.inf
            stir = np.hstack((stir, new_c))
            print(m)
            for n in range(K,J):
                if m > n:
                    continue
                elif m == n:
                    stir[n, m] = 0

                else:
                    stir[n,m] = logsumexp( [ stir[n-1, m-1] , np.log(n-1) + stir[n-1, m] ] )
            x += 1

        #np.save('stirling.npy', stir)
        #np.load('stirling.npy',)
        return stir


if __name__ == '__main__':
    stirlg = lookup_stirling(fn = 'stirling.pk')

    # Test
    ##a = stirlg.run(k = 1000)
    #b = stirlg.load()
    b = stirlg.recursive_row()
    #try:
    #    test1 = np.array_equal(a,b)
    #    print 'Test1 succesfully passed: %s' % (test1)
    #except:
    #    pass

    #for n in range(stirlg.k_max):
    #    test2 = True
    #    for m in range(stirlg.k_max):
    #        t = sym.log(stirling(n, m, kind=1)).evalf()
    #        if t == sym.zoo:
    #            t = np.inf

    #        if t != b[n,m] :
    #            print n,m, t, b[n, m]
    #            test2 = False
    #try:
    #    print 'Test2 succesfully passed: %s' % (test2)
    #except:
    #    pass
    print(b.shape)



import time
import numpy as np
from numpy import ma
from pymake import GramExp, ExpeFormat



USAGE = """\
----------------
Fit the data : This script is part of the repo/ml of pymake.
----------------
"""



class Fit(ExpeFormat):

    def __call__(self):
        return self.fit()

    def fit(self):
        expe = self.expe
        t0 = time.time()

        # Load data
        frontend = self.load_frontend()

        ### @Debug: Obsolete / Inside model
        alpha = expe.get('alpha', .1)
        gmma = expe.get('gmma', .1)
        delta = expe.get('delta', (0.5, 0.5))
        hyperparams = {'alpha': alpha, 'delta': delta, 'gmma': gmma}
        expe['hyperparams'] = hyperparams
        #############################################################

        # Load Model
        model = self.load_model(frontend)

        # Fit
        model.fit(frontend)

        self.log.info('Expe %d finished in %.1f' % (self.pt['expe']+1, time.time()-t0))


    def fitw(self):
        if self._it == 0:
            self.gramexp.check_format()

        self.expe['_write'] = True
        self()


    def fit_missing(self, ext=None):

        is_fitted = self.gramexp.make_output_path(self.expe, ext=ext, status='f')
        if not is_fitted:
            self.fitw()
            #print(self.output_path)
        else:
            self.log.info("Expe `%s' already fitted, passing..." % self._it)

    def fit_incomplete(self, ext='inf'):

        is_fitted = self.gramexp.make_output_path(self.expe, ext=ext, status='f')
        _file = self.expe['_output_path'] +'.' + ext
        try:
            if not is_fitted:
                flag = False
            else:
                flag = list(filter(None,open(_file).read().split('\n')))[-1].split()[-1]

            is_incomplete = flag != 'terminated'
        except FileNotFoundError as e:
            is_incomplete = True

        if is_incomplete:
            self.fitw()
            #print(self.output_path)
        else:
            self.log.info("Expe `%s' completed, passing..." % self._it)


if __name__ == '__main__':
    GramExp.zymake().pymake(Fit)



from collections import defaultdict
import numpy as np
import scipy as sp
from numpy import ma
import scipy.stats

from .modelbase import ModelBase

import graph_tool as gt
from graph_tool import inference


class SbmBase(ModelBase):
    __abstractmethods__ = 'model'

    def _init_params(self, frontend):
        frontend = self.frontend

        # Save the testdata
        self.data_test = frontend.data_test_w

        # For fast computation of bernoulli pmf.
        self._w_a = self.data_test[:,2].T.astype(int)
        self._w_a[self._w_a > 0] = 1
        self._w_a[self._w_a == 0] = -1
        self._w_b = np.zeros(self._w_a.shape, dtype=int)
        self._w_b[self._w_a == -1] = 1

        _len = {}
        _len['K'] = self.expe.get('K')
        _len['N'] = frontend.num_nodes()
        _len['E'] = frontend.num_edges()
        _len['nnz'] = frontend.num_nnz()
        #_len['nnz_t'] = frontend.num_nnz_t()
        _len['dims'] = frontend.num_neighbors()
        _len['nnz_ones'] = frontend.num_edges()
        _len['nnzsum'] = frontend.num_nnzsum()
        self._len = _len

        self._K = self._len['K']
        self._is_symmetric = frontend.is_symmetric()


    def _equilibrate(self):
        K = self.expe.K

        g = self.frontend.data

        measures = defaultdict(list)
        def collect_marginals(s):
            measures['entropy'].append(s.entropy())

        # Model
        state = BlockState(g, B=K)
        mcmc_equilibrate(state, callback=collect_marginals)
        #mcmc_equilibrate(state, callback=collect_marginals, force_niter=expe.iterations, mcmc_args=dict(niter=1), gibbs=True)
        #mcmc_equilibrate(state, force_niter=expe.iterations, callback=collect_marginals, wait=0, mcmc_args=dict(niter=10))

        entropy = np.array(measures['entropy'])

        #print("Change in description length:", ds)
        #print("Number of accepted vertex moves:", nmoves)

        # state.get_edges_prob(e)
        #theta = state.get_ers()

        print('entropy:')
        print(entropy)
        print(len(entropy))

    def _reduce_latent(self):
        if hasattr(self, 'get_overlap_blocks'):
            raise NotImplementedError
        else:
            try:
               _theta = self._state.get_blocks().a
               theta = np.zeros((len(_theta), self._K))
               theta[(np.arange(len(_theta)), _theta)] = 1
            except AttributeError as e:
                return self._theta, self._phi

        phi = self._state.get_matrix().A

        nr = self._state.get_nr()
        norm = np.outer(nr.a, nr.a)

        if not getattr(self, '_weighted', False):
            #phi = phi / phi.sum()
            phi = phi / norm

        self._theta, self._phi = theta, phi

        return theta, phi

    def likelihood(self, theta=None, phi=None):
        if theta is None:
            theta = self._theta
        if phi is None:
            phi = self._phi

        qijs = []
        for i,j, xij in self.data_test:
            #try:
            #   # Too long !
            #    pij = np.exp(self._state.get_edges_prob([(i,j),]))
            #except AttributeError:
            #    # If recomputed with fig :roc...
            #    pij = theta[i].dot(phi).dot(theta[j])
            pij = theta[i].dot(phi).dot(theta[j])

            qijs.append( pij )

        qijs = ma.masked_invalid(qijs)
        return qijs


    def compute_entropy(self, theta=None, phi=None, **kws):
        if 'likelihood' in kws:
            pij = kws['likelihood']
        else:
            if theta is None:
                theta, phi = self._reduce_latent()
            pij = self.likelihood(theta, phi)

        ll = pij * self._w_a + self._w_b

        ll[ll<=1e-300] = 1e-300
        # Log-likelihood
        ll = np.log(ll).sum()
        # Perplexity is 2**H(X).
        return ll
        #return self._state.entropy()

    def compute_roc(self, theta=None, phi=None, **kws):
        from sklearn.metrics import roc_curve, auc, precision_recall_curve

        if 'likelihood' in kws:
            pij = kws['likelihood']
        else:
            if theta is None:
                theta, phi = self._reduce_latent()
            pij = self.likelihood(theta, phi)

        weights = np.squeeze(self.data_test[:,2].T)

        y_true = weights.astype(bool)*1
        self._probas = pij
        self._y_true = y_true

        fpr, tpr, thresholds = roc_curve(y_true, pij)
        roc = auc(fpr, tpr)
        return roc

    def compute_pr(self, *args, **kwargs):
        from sklearn.metrics import average_precision_score
        return average_precision_score(self._y_true, self._probas)

    def compute_wsim(self, *args, **kws):
        return None


    def _spec_from_expe(self, _model):
        ''' Set Sklearn parameters. '''
        import inspect # @temp to be integrated

        model_params = list(inspect.signature(_model).parameters)
        spec = dict()
        spec_map = getattr(self, 'spec_map', {})
        default_spec = getattr(self, '_default_spec', {})
        for k in model_params:
            if k in list(self.expe)+list(spec_map):
                _k = spec_map.get(k, k)
                if _k in self.expe:
                    spec[k] = self.expe[_k]
                elif callable(_k):
                    spec[k] = _k(self)
            elif k in default_spec:
                spec[k] = default_spec[k]


        return spec

    def fit(self, frontend):
        self._init(frontend)
        g = self.frontend.data

        fit_fun = inference.minimize_blockmodel_dl
        spec = self._spec_from_expe(fit_fun)

        self.log.info("Fitting `%s' model with spec: %s" % (type(self), str(spec)))
        self._state = fit_fun(g, **spec)

        #inference.mcmc_equilibrate(self._state)

        self.compute_measures()
        if self.expe.get('_write'):
            self.write_current_state(self)


class SBM_gt(SbmBase):

    _default_spec = dict(deg_corr=False, overlap=False, layers=False)
    spec_map = dict(B_min='K', B_max='K')


class OSBM_gt(SbmBase):

    _default_spec = dict(deg_corr=False, overlap=True, layers=False)
    spec_map = dict(B_min='K', B_max='K')

class WSBM_gt(SbmBase):

    _default_spec = dict(deg_corr=False, overlap=False, layers=False)
    spec_map = dict(B_min='K', B_max='K', state_args=lambda self: {'recs':[self.frontend.data.ep.weights], 'rec_types' : ["discrete-poisson"]})
    _weighted = True

    def compute_entropy(self, theta=None, phi=None, **kws):
        if 'likelihood' in kws:
            pij = kws['likelihood']
        else:
            if theta is None:
                theta, phi = self._reduce_latent()
            pij = self.likelihood(theta, phi)

        weights = self.data_test[:,2].T
        ll = sp.stats.poisson.pmf(weights, pij)

        ll[ll<=1e-300] = 1e-300
        # Log-likelihood
        ll = np.log(ll).sum()
        # Perplexity is 2**H(X).
        return ll
        #return self._state.entropy()

    def compute_roc(self, theta=None, phi=None, treshold=1, **kws):
        from sklearn.metrics import roc_curve, auc, precision_recall_curve

        if 'likelihood' in kws:
            pij = kws['likelihood']
        else:
            if theta is None:
                theta, phi = self._reduce_latent()
            pij = self.likelihood(theta, phi)

        trsh = treshold
        weights = np.squeeze(self.data_test[:,2].T)

        nr = self._state.get_nr()
        norm = np.outer(nr.a, nr.a)
        phi = phi / norm
        #phi = phi / phi.sum()

        probas = []
        for i,j, xij in self.data_test:
            pij =  theta[i].dot(phi).dot(theta[j])
            probas.append( pij )

        probas = ma.masked_invalid(probas)
        #probas = 1 - sp.stats.poisson.pmf(0, probas)
        #probas = pij

        y_true = weights.astype(bool)*1
        self._y_true = y_true
        self._probas = probas

        fpr, tpr, thresholds = roc_curve(y_true, probas)
        roc = auc(fpr, tpr)
        return roc

    def compute_wsim(self, theta=None, phi=None, **kws):
        if 'likelihood' in kws:
            pij = kws['likelihood']
        else:
            if theta is None:
                theta, phi = self._reduce_latent()
            pij = self.likelihood(theta, phi)

        weights = self.data_test[:,2].T

        nr = self._state.get_nr()

        ws = np.array([ theta[i].dot(phi).dot(theta[j]) for i,j,w in self.data_test if w > 0])

        # l1 norm
        wd = weights[weights>0]
        nnz = len(wd)
        mean_dist = np.abs(ws - wd).sum() / nnz
        return mean_dist

class WSBM2_gt(SbmBase):
    # graph_tool work in progress for weithed networds...
    _default_spec = dict(deg_corr=False, overlap=False, layers=False)
    spec_map = dict(B_min='K', B_max='K',
                    state_args=lambda self:{'eweight': self.frontend.data.ep.weights,'recs':[self.frontend.data.ep.weights], 'rec_types': ["discrete-poisson"]})




class OWSBM_gt(SbmBase):
    pass





# -*- coding: utf-8 -*-

from time import time
import itertools
import logging
lgg = logging.getLogger('root')

import numpy as np
import scipy as sp
from numpy import ma
from scipy.sparse import csr_matrix
from scipy.sparse import lil_matrix
sp_dot = csr_matrix.dot

from .ibp import IBP
from ml.model.modelbase import GibbsSampler
from pymake.util.algo import *

# We will be taking log(0) = -Inf, so turn off this warning
#np.seterr(divide='ignore')

"""
Implements MCMC inference for the Infinite Latent Feature Relationnal Model [1].
This code was modified from the code originally written by Zhai Ke (kzhai@umd.edu).

[1] Kurt Miller, Michael I Jordan, and Thomas L Griffiths. Nonparametric latent feature models for link prediction. In Advances in neural information processing systems 2009.
"""

### @TODO
#   * 2 parameter ibp
#   * poisson-gamma ibp for real valued relation
#   * Oprimization:
#   *   - updating only the subset of relation affected by feature modified (active in both sides).
#   * Structure:
#       * Sample methods, divide and conquer
#

W_diag = -2

class IBPGibbsSampling(IBP, GibbsSampler):
    __abstractmethods__ = 'model'
    def __init__(self, expe, frontend,
                 assortativity=False,
                 alpha_hyper_parameter=None,
                 sigma_w_hyper_parameter=None,
                 metropolis_hastings_k_new=True,):
        self._sigma_w_hyper_parameter = sigma_w_hyper_parameter
        self.bilinear_matrix = None
        self.log_likelihood = None
        self.assortativity = assortativity
        self._overflow = 1.0
        self.ratio_MH_F = 0.0
        self.ratio_MH_W = 0.0
        self.snapshot_freq = 20

        self.burnin = expe.get('burnin',  5) # (inverse burnin, last sample to keep
        self.thinning = expe.get('thinning',  1)
        self._csv_typo = '_iteration time_it _entropy _entropy_t _K _alpha _sigma_w Z_sum ratio_MH_F ratio_MH_W'
        #self._fmt = '%d %.4f %.8f %.8f %d %.8f %.8f %d %.4f %.4f'
        IBP.__init__(self, alpha_hyper_parameter, metropolis_hastings_k_new)
        GibbsSampler.__init__(self, expe, frontend)
    """
    @param data: a NxD np data matrix
    @param alpha: IBP hyper parameter
    @param sigma_w: standard derivation of the feature
    @param initialize_Z: seeded Z matrix """
    def _initialize(self, frontend, alpha=1.0, sigma_w=1,
                    initial_Z=None, initial_W=None, KK=None):

        self._mean_w = 0
        assert(type(sigma_w) is float)
        self._sigma_w = sigma_w
        self._sigb = 1 # Carreful make overflow in exp of sigmoid !

        if frontend is None:
            return

        self.mask = frontend.data_ma.mask

        self.symmetric = frontend.is_symmetric()
        self.nnz = len(frontend.data_ma.compressed())
        super(IBPGibbsSampling, self)._initialize(frontend, alpha, initial_Z, KK=KK)

        self._W_prior = np.zeros((1, self._K))
        if initial_W != None:
            self._W = initial_W
        else:
            if self.assortativity == 1:
                # Identity
                self._W  = (np.ones((self._K, self._K))*W_diag) * (np.ones((self._K)) + np.eye(self._K)*-2)
            elif self.assortativity == 2:
                # Bivariate Gaussian
                v = 10
                x, y = np.mgrid[-v:v:self._K*1j, -v:v:self._K*1j]
                xy = np.column_stack([x.flat, y.flat])
                mu = np.array([0, 0])
                sigma = np.array([1, 1])
                covariance = np.array([[v*100,0],[0,v/10]])
                theta = np.pi / 4
                rot = np.array([[np.cos(theta), -np.sin(theta)],[np.sin(theta), np.cos(theta)]])
                covariance = rot.dot(covariance).dot(rot.T)
                z = sp.stats.multivariate_normal.pdf(xy, mean=mu, cov=covariance)
                z = 400 * z.reshape(x.shape)

                self.z_mean = z - np.ones(z.shape)*1
                self._W = np.random.normal(self.z_mean, self._sigma_w, (self._K, self._K))
            else:
                self._W = np.random.normal(self._mean_w, self._sigma_w, (self._K, self._K))

            if self.symmetric:
                self._W = np.tril(self._W) + np.tril(self._W, -1).T
                np.fill_diagonal(self._W, 1)

        #self._Z = csr_matrix(self._Z)
        #self._Z = lil_matrix(self._Z)

        assert(self._W.shape == (self._K, self._K))

    def fit(self, *args, **kwargs):
        # Two things to merge unify with GibbsSample !!!
        #   * rename parameters _Z and _W ot _theta and _Phi
        #   * appand all the samplee in self.s to factorize the method self.sample

        self._init()

        lgg.info( '__init__  Entropy: %f' % (-self.log_likelihood_Y() / self.nnz))
        for _it in range(self.iterations):
            self._iteration = _it

            begin_it = time()
            self.sample()

            if _it >= self.iterations - self.burnin:
                if _it % self.thinning == 0:
                    self.samples.append([self._Z, self._W])

            self.compute_measures(begin_it)
            lgg.info("iteration: %i,  Entropy : %f \t\t K=%i,  Entropy Z: %f, alpha: %f sigma_w: %f Z.sum(): %i" % (_it, self._entropy, self._K, self._entropy_Z, self._alpha, self._sigma_w, self.Z_sum))
            if self._write:
                self.write_current_state(self)
                if _it > 0 and _it % self.snapshot_freq == 0:
                    self.save(silent=True)

        ### Clean Things
        if not self.samples:
            self.samples.append([self._Z, self._W])
        self._reduce_latent()
        self.samples = None # free space

        Yd = self._Y.data
        Yd[Yd <= 0 ] = 0
        Yd[Yd > 0 ] = 1

        return

    def compute_measures(self, begin_it=0):

        ### Output / Measures
        self._entropy = self.log_likelihood
        self._entropy_t = None
        self._entropy_Z = np.nan # self.log_likelihood_Z()
        self.Z_sum = (self._Z == 1).sum()

        self.time_it = time() - begin_it

    def sample(self):

        # Can't get why I need this !
        self.log_likelihood_Y()
        # Sample every object
        order = np.random.permutation(self._N)
        for (object_counter, object_index) in enumerate(order):
            #sys.stdout.write('Z')
            singleton_features = self.sample_Zn(object_index)

            if self._metropolis_hastings_k_new:
                if self.metropolis_hastings_K_new(object_index, singleton_features):
                    #sys.stdout.write('Z+')
                    self.ratio_MH_F += 1
            #sys.stdout.flush()

        self.ratio_MH_F /= len(order)

        # Regularize matrices
        self.regularize_matrices()

        if self.assortativity == 1:
            self._W  = (np.ones((self._K, self._K))*W_diag) * (np.ones((self._K)) + np.eye(self._K)*-2)
            #self._W  = np.eye(self._K)
        elif self.assortativity == 2:
            self.sample_W()
        else:
            self.sample_W()

        if self._alpha_hyper_parameter:
            self._alpha = self.sample_alpha()

        if self._sigma_w_hyper_parameter != None:
            self._sigma_w = self.sample_sigma_w(self._sigma_w_hyper_parameter)

    def sample_Zn(self, object_index):
        ''' :param: object_index: an int data type, indicates the object index (row index) of Z we want to sample '''

        assert(type(object_index) == int or type(object_index) == np.int32 or type(object_index) == np.int64)

        # calculate initial feature possess counts
        m = self._Z.sum(axis=0)

        # remove this data point from m vector
        new_m = (m - self._Z[object_index, :]).astype(np.float)

        #m = np.array(m).reshape(-1)
        #new_m = np.array(new_m).reshape(-1)

        # compute the log probability of p(Znk=0 | Z_nk) and p(Znk=1 | Z_nk)
        log_prob_z0 = np.log(1.0 - new_m / self._N)
        log_prob_z1 = np.log(new_m / self._N)
        log_prob_z = {0: log_prob_z0, 1: log_prob_z1}

        # find all singleton features possessed by current object
        singleton_features = [nk for nk in range(self._K) if self._Z[object_index, nk] != 0 and new_m[nk] == 0]
        non_singleton_features = [nk for nk in range(self._K) if nk not in singleton_features]

        order = np.random.permutation(self._K)
        for (feature_counter, feature_index) in enumerate(order):
            if feature_index in non_singleton_features:
                old_znk = self._Z[object_index, feature_index]
                new_znk = 1 if old_znk == 0 else 0

                # compute the log likelihood when Znk = old
                #log_old = self.log_likelihood_Y(object_index=(object_index, feature_index))
                log_old = self.log_likelihood
                bilinear = self.bilinear_matrix.copy()
                #log_old_1 = self.log_likelihood_Y(object_index=(object_index, feature_index))
                #if (not (bilinear == self.bilinear_matrix).all()):
                #    print feature_counter, object_index, feature_index
                #    print np.where(self.bilinear_matrix != bilinear)
                self._overflow = - log_old -1
                prob_old = log_old + log_prob_z[old_znk][feature_index]

                # compute the log likelihood when Znk = new
                self._Z[object_index, feature_index] = new_znk
                log_new = self.log_likelihood_Y(object_index=(object_index, feature_index))
                if log_new > -self._overflow:
                    self._overflow = - log_new -1
                prob_new = log_new + log_prob_z[new_znk][feature_index]

                prob_new = np.exp(prob_new + self._overflow)
                prob_old = np.exp(prob_old + self._overflow)
                Znk_is_new = prob_new / (prob_new + prob_old)
                if Znk_is_new > 0 and np.random.random() <= Znk_is_new:
                    # gibbs_accept ++
                    pass
                else:
                    self._Z[object_index, feature_index] = old_znk
                    self.bilinear_matrix = bilinear
                    self.log_likelihood = log_old

        return singleton_features

    """
    sample K_new using metropolis hastings algorithm """
    def metropolis_hastings_K_new(self, object_index, singleton_features):
        if type(object_index) != list:
            object_index = [object_index]

        # sample K_new from the metropolis hastings proposal distribution, i.e., a poisson distribution with mean \frac{\alpha}{N}
        K_temp = sp.stats.poisson.rvs(self._alpha / self._N)
        K_new = self._K + K_temp - len(singleton_features)

        if K_new <= 0 or K_temp <= 0 and len(singleton_features) <= 0:
            return False

        # generate new weight from a normal distribution with mean 0 and variance sigma_w, a K_new-by-K matrix
        non_singleton_features = [k for k in range(self._K) if k not in singleton_features]
        W_temp_v = np.random.normal(self._mean_w, self._sigma_w, (K_temp, self._K - len(singleton_features)))
        W_temp_h = np.random.normal(self._mean_w, self._sigma_w, (K_new, K_temp))
        W_new = np.delete(self._W, singleton_features,0)
        W_new = np.vstack((np.delete(W_new, singleton_features,1), W_temp_v))
        W_new = np.hstack((W_new, W_temp_h))
        # generate new features z_i row
        Z_new_r = np.hstack((self._Z[[object_index], non_singleton_features], np.ones((len(object_index), K_temp))))
        #print K_temp, object_index, non_singleton_features

        # compute the probability of generating new features / old featrues
        log_old = self.log_likelihood
        Z_new = np.hstack((self._Z[:, non_singleton_features], np.zeros((self._N, K_temp))))
        Z_new[object_index, :] = Z_new_r
        bilinear_matrix = self.bilinear_matrix.copy()
        log_new = self.log_likelihood_Y(Z=Z_new, W=W_new)
        self._overflow = - log_new -1
        prob_new = np.exp(log_new + self._overflow)
        prob_old = np.exp(log_old + self._overflow)

        assert(W_new.shape == (K_new, K_new))
        assert(Z_new.shape == (self._N, K_new))

        # compute the probability of generating new features
        r = prob_new / prob_old
        MH_accept = min(1, r)

        # if we accept the proposal, we will replace old W and Z matrices
        if np.random.random() <= MH_accept and not np.isnan(r):
            # construct W_new and Z_new
            #print 'MH_accept: %s, singleton feature: %s, k_new: %s' % (MH_accept, len(singleton_features), K_temp)
            self._W = W_new
            self._Z = Z_new
            self._K = K_new
            return True
        else:
            self.bilinear_matrix = bilinear_matrix
            self.log_likelihood = log_old
            return False

    """
    Sample W using metropolis hasting """
    def sample_W(self):
        #sys.stdout.write('W')
        # sample every weight
        sigma_rw = 1.0
        if self.symmetric:
            order = np.arange(self._K**2).reshape(self._W.shape)
            iu = np.triu_indices(self._K)
            order = np.random.permutation(order[iu])
        else:
            order = np.random.permutation(self._K**2)
        for (observation_counter, observation_index) in enumerate(order):
            w_old = self._W.flat[observation_index]
            w_new = np.random.normal(w_old, sigma_rw)
            j_new = sp.stats.norm(w_old, sigma_rw).pdf(w_new)
            j_old = sp.stats.norm(w_new, sigma_rw).pdf(w_old)
            if self.assortativity == 2:
                mean = self.z_mean.flat[observation_index]
            else:
                mean = 0
            pw_new = sp.stats.norm(mean, self._sigma_w).pdf(w_new)
            pw_old = sp.stats.norm(mean, self._sigma_w).pdf(w_old)

            log_old = self.log_likelihood
            bilinear_matrix = self.bilinear_matrix.copy()
            self._overflow = - log_old -1
            self._W.flat[observation_index] = w_new
            if self.symmetric:
                indm = np.unravel_index(observation_index, self._W.shape)
                self._W.T[indm] = w_new
            log_new = self.log_likelihood_Y(object_index=observation_index)
            if -log_new < self._overflow:
                self._overflow = - log_new -1
            likelihood_new = np.exp(log_new + self._overflow)
            likelihood_old = np.exp(log_old + self._overflow)
            r = likelihood_new * pw_new * j_old / ( likelihood_old * pw_old * j_new )
            MH_accept = min(1, r)

            if np.random.random() <= MH_accept and not np.isnan(r):
                self.ratio_MH_W += 1
            else:
                self.bilinear_matrix = bilinear_matrix
                self.log_likelihood = log_old
                self._W.flat[observation_index] = w_old
                if self.symmetric:
                    self._W.T[indm] = w_old

        try:
            self.ratio_MH_W /= len(order)
        except:
            pass
        return self.ratio_MH_W

    """
    sample feature variance, i.e., sigma_w """
    def sample_sigma_w(self, sigma_w_hyper_parameter):
        return self.sample_sigma(self._sigma_w_hyper_parameter, self._W - np.tile(self._W_prior, (self._K, 1)))

    """
    remove the empty column in matrix Z and the corresponding feature in W """
    def regularize_matrices(self):
        Z_sum = np.sum(self._Z, axis=0)
        indices = np.nonzero(Z_sum == 0)

        if 0 in Z_sum:
            lgg.warn( "need to regularize matrices, feature to all zeros !")

        #self._Z = self._Z[:, [k for k in range(self._K) if k not in indices]]
        #self._W = self._W[[k for k in range(self._K) if k not in indices], :]
        #self._K = self._Z.shape[1]
        #assert(self._Z.shape == (self._N, self._K))
        #assert(self._W.shape == (self._K, self._K))

    """
    compute the log-likelihood of the data Y """
    def log_likelihood_Y(self, Z=None, W=None, object_index=None):
        if W is None:
            W = self._W
        if Z is None:
            Z = self._Z

        (N, K) = Z.shape
        assert(W.shape == (K, K))

        bilinear_init = self.bilinear_matrix is not None
        #bilinear_init = False

        if type(object_index) is tuple and bilinear_init:
            # Z update
            n = object_index[0]
            k = object_index[1]
            self.bilinear_matrix[n,:] = self.logsigmoid(Z[n].dot(W).dot( Z.T ), self._Y[n,:])
            #self.bilinear_matrix[n,:] = self.logsigmoid( Z.dot(Z[n].dot(W).T).T , self._Y[n,:])
            if self.symmetric:
                self.bilinear_matrix[:,n] = self.bilinear_matrix[n,:]
            else:
                self.bilinear_matrix[:,n] = self.logsigmoid(Z.dot(W).dot(Z[n]), self._Y[:, n])
        elif np.issubdtype(object_index, np.integer) and bilinear_init:
            # W update
            indm = np.unravel_index(object_index, W.shape)
            ki, kj = indm
            sublinear = list( np.where(Z[:, ki] > 0)[0])
            sublinear = sorted(list(set(sublinear + list(np.where(Z[:, kj] > 0)[0]) )))
            if len(sublinear) > 0:
                self.bilinear_matrix[np.ix_(sublinear, sublinear)] = self.logsigmoid( Z[sublinear].dot(W).dot(Z[sublinear].T), self._Y[np.ix_(sublinear, sublinear)] )
                #self.bilinear_matrix[np.ix_(sublinear, sublinear)] = self.logsigmoid( sp_dot(Z[sublinear].dot(W), Z[sublinear].T), Y[np.ix_(sublinear, sublinear)] )
        else:
            # Check speed with sparse matrix here !
            self.bilinear_matrix = self.logsigmoid(Z.dot(W).dot(Z.T))
            #self.bilinear_matrix = self.logsigmoid( Z.dot(Z.dot(W).T).T )

        self.log_likelihood = np.sum(self.bilinear_matrix)
        if np.abs(self.log_likelihood) == np.inf:
            print("dohn debug me here")
            if self._sigb >= 2:
                self._sigb -= 1
            else:
                self._sigb = 1
                self._W /= 2
                W = self._W
            self.bilinear_matrix = self.logsigmoid(Z.dot(W).dot(Z.T))
            self.log_likelihood = np.sum(self.bilinear_matrix)

        return self.log_likelihood

    def logsigmoid(self, X, Y=None):
        if Y is None:
            Y = self._Y
        # 1 - sigmoid(x) = sigmoid(-x) ~ Y * ...
        v = - np.log(1 + np.exp(- Y * self._sigb * X))
        return v

    """
    compute the log-likelihood of W """
    def log_likelihood_W(self):
        log_likelihood = - 0.5 * self._K * self._D * np.log(2 * np.pi * self._sigma_w * self._sigma_w)
        #for k in range(self._K):
        #    W_prior[k, :] = self._mean_a[0, :]
        W_prior = np.tile(self._W_prior, (self._K, 1))
        log_likelihood -= np.trace(np.dot((self._W - W_prior).transpose(), (self._W - W_prior))) * 0.5 / (self._sigma_w ** 2)

        return log_likelihood

    """
    compute the log-likelihood of the model """
    def log_likelihood_model(self):
        # loglikelihood_W ? Z ?
        return self.log_likelihood_Y()

    # @ibp
    def update_hyper(self, hyper):
        if hyper is None:
            return
        elif isinstance(type(hyper), (tuple, list)):
            alpha = hyper[0]
        else:
            alpha = hyper.get('alpha')

        if alpha:
            self._alpha = alpha

    # @ibp
    def get_hyper(self):
        alpha = self._alpha
        delta = (self._mean_w, self._sigma_w)
        return (alpha, delta)

    # @ibp
    def generate(self, N, K=None, nodelist=None, hyperarams=None, mode='predictive', symmetric=True, **kwargs):
        self.update_hyper(hyperarams)
        alpha, delta = self.get_hyper()
        if mode == 'generative':
            N = int(N)

            # Use for the stick breaking generation
            #K = alpha * np.log(N)

            # Generate F
            theta = self.initialize_Z(N, alpha)
            K = theta.shape[1]

            # Generate Phi
            phi = np.random.normal(delta[0], delta[1], size=(K,K))
            if symmetric is True:
                phi = np.triu(phi) + np.triu(phi, 1).T

            self._theta = theta
            self._phi = phi
        elif mode == 'predictive':
            theta, phi = self.get_params()
            K = theta.shape[1]

        if nodelist:
            Y = Y[nodelist, :][:, nodelist]
            phi = phi[nodelist, :]

        likelihood = self.likelihood(theta, phi)
        #likelihood[likelihood >= 0.5 ] = 1
        #likelihood[likelihood < 0.5 ] = 0
        #Y = likelihood
        Y = sp.stats.bernoulli.rvs(likelihood)
        return Y, theta, phi

    def likelihood(self, theta=None, phi=None):
        if theta is None:
            try:
                theta = self._theta
            except:
                theta = self._Z
        if phi is None:
            try:
                phi = self._phi
            except:
                phi = self._W
        bilinear_form = theta.dot(phi).dot(theta.T)
        likelihood = 1 / (1 + np.exp(- self._sigb * bilinear_form))
        return likelihood

    def _reduce_latent(self):
        Z, W = list(map(list, zip(*self.samples)))
        ks = [ mat.shape[1] for mat in Z]
        bn = np.bincount(ks)
        k_win = np.argmax(bn)
        lgg.debug('K selected: %d' % k_win)

        ind_rm = []
        [ind_rm.append(i) for i, v in enumerate(Z) if v.shape[1] != k_win]
        for i in sorted(ind_rm, reverse=True):
            Z.pop(i)
            W.pop(i)

        lgg.debug('Samples Selected: %d over %s' % (len(Z), len(Z)+len(ind_rm) ))

        Z = Z[-1]
        W = np.mean(W, 0)
        self._theta = Z
        self._phi = W
        self._K = self._theta.shape[1]
        return Z, W

    def get_clusters(self, K=None):
        Z, W = self.get_params()
        K = K or Z.shape[1]
        Z = self.leftordered(Z)
        clusters = kmeans(Z, K=K)
        return clusters

    def _get_clusters(self, K=None):
        Z, W = self.get_params()
        K = K or Z.shape[1]
        clusters = np.argmax(Z * np.tile(W.sum(0), (Z.shape[0],1)), 1)
        return clusters

    # add sim optionsin clusters
    def get_communities(self, K=None):
        Z, W = self.get_params()
        K = K or Z.shape[1]
        Z = self.leftordered(Z)
        clusters = kmeans(Z.dot(W), K=K)
        return clusters

    #@wrapper !
    def communities_analysis(self, data, clustering='modularity'):
        symmetric = (data == data.T).all()

        # @debug !!!
        clusters = self.get_clusters()
        #nodes_list = [k[0] for k in sorted(zip(range(len(clusters)), clusters), key=lambda k: k[1])]
        Z, W = self.get_params()
        block_hist = Z.sum(0)

        local_degree = {}
        if symmetric:
            k_perm = np.unique( list(map(list, map(set, itertools.product(np.unique(clusters) , repeat=2)))))
        else:
            k_perm = itertools.product(np.unique(clusters) , repeat=2)

        for c in k_perm:
            if type(c) in (np.float64, np.int64):
                # one clusters (as it appears for real with max assignment
                l = k = c
            elif  len(c) == 2:
                # Stochastic Equivalence (extra class bind
                k, l = c
            else:
                # Comunnities (intra class bind)
                k = l = c.pop()
            comm = '-'.join([str(k), str(l)])
            local = local_degree.get(comm, [])

            C = np.tile(clusters, (data.shape[0],1))
            y_c = data * ((C==k) & (C.T==l))
            if y_c.size > 0:
                local_degree[comm] = list(adj_to_degree(y_c).values())

        self.comm = {'local_degree':local_degree,
                     'clusters':clusters,
                     'block_hist': block_hist,
                     'size': len(block_hist)}

        return self.comm

    def blockmodel_ties(self, data, remove_empty=True):
        """ return ties based on a weighted average
            of the local degree ditribution """

        if not hasattr(self, 'comm'):
            self.communities_analysis(data)
        comm = self.comm

        m = [ wmean(a[0], a[1], mean='arithmetic')  for d in comm['local_degree'].values() for a in [degree_hist(d)] if len(a[0])>0 ]
        # Mean
        #m = map(np.mean, comm['local_degree'].values())
        # Variance Unused How to represant that !?
        #v = np.array(map(np.std, comm['local_degree'].values()))

        # factorize by using clusters hist instead
        hist, label = sorted_perm(m, comm['local_degree'].keys(), reverse=True)

        if remove_empty is True:
            null_classes = (hist == 0).sum()
            if null_classes > 0:
                hist = hist[:-null_classes]; label = label[:-null_classes]

        bm = zip(label, hist)
        self.comm['block_ties'] = bm
        return bm

    def purge(self):
        purge_obj = ['frontend', '_Y']
        for obj in purge_obj:
            if hasattr(self, obj):
                setattr(self, obj, None)





#!/usr/bin/env python3
import subprocess
import os



# Test Fonctionel



test_design = {
    '../repo/ml':[
        'pmk',
        'pmk update',
        'pmk show',
        'pmk -l',
        'pmk -ll',
        'pmk -l spec',
        'pmk -l model',
        'pmk -l script',
        'pmk -l model',
        'pmk default_expe',
        'pmk path default_expe',
        'pmk default_expe -x fit -w',
        'pmk default_expe -x plot',
        'pmk default_expe -x plot fig corpus:_entropy',
        'pmk hist'],
    '../docsearch': ['pmk -x search tree'],
}

n_errors = 0
n_test = sum([len(tests) for tests in test_design])

for repo, tests in test_design.items():

    os.chdir(repo)

    for test in tests:

        cmd = test

        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        print('Testing :: %s' % cmd)
        out, err = p.communicate()
        result = str(out).split('\n')

        ### Output
        if False:
            for lin in result:
                if not lin.startswith('#'):
                    print(lin)

        if p.returncode != 0:
            n_errors += 1
            print("test failed: %d,  %s" % (p.returncode, err))

print('Test Sucess: %d / %d' % (n_test-n_errors, n_test))




import os
from pymake import ExpeFormat

from collections import OrderedDict

import matplotlib.pyplot as plt
from pymake.util.utils import colored



USAGE = """\
----------------
Manage the data : This script is part of the repo/ml of pymake.
----------------
"""



class Data(ExpeFormat):

    _default_expe = { '_expe_silent' : True }

    # @need an expe
    def completed(self, ext=None):
        ''' Show completed expe. '''
        if self.is_first_expe():
            self.D.n_exp_total = self.expe_size
            self.D.n_exp_completed = 0

        is_fitted = self.gramexp.make_output_path(self.expe, ext=ext, status='f')
        if is_fitted:
            self.D.n_exp_completed += 1
            self.log.debug(self.expe['_output_path'])


        if self.is_last_expe():
            table = OrderedDict([('completed', [self.D.n_exp_completed]),
                                 ('total', [self.D.n_exp_total]),
                                ])
            print (self.tabulate(table, headers='keys', tablefmt='simple', floatfmt='.3f'))


    # @need an expe
    def missing(self, ext=None):
        ''' Show missing expe. '''
        if self.is_first_expe():
            self.D.n_exp_total = self.expe_size
            self.D.n_exp_missing = 0

        is_fitted = self.gramexp.make_output_path(self.expe, ext=ext, status='f')
        if not is_fitted:
            self.D.n_exp_missing += 1
            self.log.debug(self.expe['_output_path'])


        if self.is_last_expe():
            table = OrderedDict([('missing', [self.D.n_exp_missing]),
                                 ('total', [self.D.n_exp_total]),
                                ])
            print (self.tabulate(table, headers='keys', tablefmt='simple', floatfmt='.3f'))


    # @need an expe
    def incomplete(self, ext='inf'):
        ''' Show not terminated expe. '''
        if self.is_first_expe():
            self.D.n_exp_total = self.expe_size
            self.D.n_exp_missing = 0

        is_fitted = self.gramexp.make_output_path(self.expe, ext=ext, status='f')
        _file = self.expe['_output_path'] +'.' + ext
        try:
            if not is_fitted:
                flag = False
            else:
                flag = list(filter(None,open(_file).read().split('\n')))[-1].split()[-1]

            is_incomplete = flag != 'terminated'
        except FileNotFoundError as e:
            is_incomplete = True

        if is_incomplete:
            self.D.n_exp_missing += 1
            self.log.debug(self.expe['_output_path'])


        if self.is_last_expe():
            table = OrderedDict([('incomplete', [self.D.n_exp_missing]),
                                 ('total', [self.D.n_exp_total]),
                                ])
            print (self.tabulate(table, headers='keys', tablefmt='simple', floatfmt='.3f'))

    def topo(self):
        print('''' Todo Topo:

                1. get all _typo in spec,
                2. parse all file (pk ou inf ?)
                3. classify all expe accordinf to : refdir, _name, corpus, model
                4. tabulate.
             ''')
        pass

    def move(self, *args, copy=False):
        ''' move a experiences files (all extension)
            (simulta by default. Use -f to force the operation)

            *args are sequence of parameters to change in the filename with the syntax:
                -x move foo=bar
            which means that the paramters {foo} in the output_path will be modified with the value {bar}
        '''
        import glob
        import shutil
        assert(len(args)>0)

        expe = self.expe
        new_expe = expe.copy()

        for o in args:
            k,v = o.split('=')
            new_expe[k] = v

        opath = self.output_path
        npath = self.gramexp.make_output_path(new_expe)

        if self.is_first_expe():
            self.log.info('Moving files for request: %s' % args)
            self.D.mesg = []
            self.D.num_expe = 0
            self.D.num_total = 0

        self.D.num_total += 1

        textm = 'copying' if copy  else 'moving'
        cwd = os.getenv('PWD')
        pwd_len = len(cwd)

        flag = False
        for ofn in glob.glob(opath+'.*'):
            ext = ofn[len(opath):]
            nfn = npath + ext
            self.D.mesg.append("%s `%s' -> `%s'" % (textm, './'+ofn[pwd_len+1:], './'+nfn[pwd_len+1:]))

            if expe.get('force'):
                if copy is True:
                    shutil.copyfile(ofn, nfn)
                else:
                    shutil.move(ofn, nfn)
                flag = True

        if flag:
            self.D.num_expe +=1

        if self.is_last_expe():
            if not expe.get('force'):
                print('*** Simulation (need force option.) ***')
            textm = 'copied' if copy else 'moved'
            self.log.debug('\n'.join(self.D.mesg)+'\n')
            self.log.info('%d/%d expe %s (%d files).' % (self.D.num_expe, self.D.num_total, textm, len(self.D.mesg)))


    def copy(self, *args):
        ''' like move but copy files instead.'''

        self.move(*args, copy=True)




# -*- coding: utf-8 -*-
"""
@author: Ke Zhai (zhaike@cs.umd.edu)

Implements uncollapsed Gibbs sampling for the linear-Gaussian infinite latent feature model (IBP).
"""

import numpy, scipy;
import math, random;
from .gs import GibbsSampling;
import scipy.stats;

# We will be taking log(0) = -Inf, so turn off this warning
numpy.seterr(divide='ignore')

class SemicollapsedGibbsSampling(GibbsSampling):
    """
    @param data: a NxD NumPy data matrix
    @param alpha: IBP hyper parameter
    @param sigma_x: standard derivation of the noise on data, often referred as sigma_n as well
    @param sigma_a: standard derivation of the feature, often referred as sigma_f as well
    @param initializ_Z: seeded Z matrix
    """
    def _initialize(self, data, alpha=1.0, sigma_f=1.0, sigma_x=1.0, initial_Z=None, A_prior=None, initial_A=None):
        # Data matrix
        super(SemicollapsedGibbsSampling, self)._initialize(self.center_data(data), alpha, sigma_f, sigma_x, A_prior, initial_Z);

        if initial_A != None:
            # this will replace the A matrix generated in the super class.
            self._A = initial_A;
        assert(self._A.shape == (self._K, self._D));

    """
    sample the corpus to train the parameters
    """
    def sample(self, iteration, directory="../../output/tmp-output/"):
        import os, shutil
        if os.path.exists(directory):
            shutil.rmtree(directory);

        assert(self._Z.shape == (self._N, self._K));
        assert(self._A.shape == (self._K, self._D));
        assert(self._X.shape == (self._N, self._D));

        #sample the total data
        for iter in xrange(iteration):
            # sample every object
            order = numpy.random.permutation(self._N);
            for (object_counter, object_index) in enumerate(order):
                # sample Z_n
                singleton_features = self.sample_Zn(object_index);

                if self._metropolis_hastings_k_new:
                    # sample K_new using metropolis hasting
                    self.metropolis_hastings_K_new(object_index, singleton_features);

            self.sample_A();

            if self._alpha_hyper_parameter != None:
                self._alpha = self.sample_alpha();

            if self._sigma_x_hyper_parameter != None:
                self._sigma_x = self.sample_sigma_x(self._sigma_x_hyper_parameter);

            if self._sigma_a_hyper_parameter != None:
                self._sigma_a = self.sample_sigma_a(self._sigma_a_hyper_parameter);

            print("iteration: %i\tK: %i\tlikelihood: %f" % (iter, self._K, self.log_likelihood_model()));
            print("alpha: %f\tsigma_a: %f\tsigma_x: %f" % (self._alpha, self._sigma_a, self._sigma_x));

            if (iter + 1) % self._snapshot_interval == 0:
                self.export_snapshot(directory, iter + 1);

    """
    @param object_index: an int data type, indicates the object index (row index) of Z we want to sample
    """
    def sample_Zn(self, object_index):
        assert(type(object_index) == int or type(object_index) == numpy.int32 or type(object_index) == numpy.int64);

        # calculate initial feature possess counts
        m = self._Z.sum(axis=0);

        # remove this data point from m vector
        new_m = (m - self._Z[object_index, :]).astype(numpy.float);

        # compute the log probability of p(Znk=0 | Z_nk) and p(Znk=1 | Z_nk)
        log_prob_z1 = numpy.log(new_m / self._N);
        log_prob_z0 = numpy.log(1.0 - new_m / self._N);

        # find all singleton features possessed by current object
        singleton_features = [nk for nk in range(self._K) if self._Z[object_index, nk] != 0 and new_m[nk] == 0];
        non_singleton_features = [nk for nk in range(self._K) if nk not in singleton_features]

        order = numpy.random.permutation(self._K);
        for (feature_counter, feature_index) in enumerate(order):
            if feature_index in non_singleton_features:
                #old_Znk = self._Z[object_index, feature_index];

                # compute the log likelihood when Znk=0
                self._Z[object_index, feature_index] = 0;
                prob_z0 = self.log_likelihood_X(self._X[[object_index], :], self._Z[[object_index], :]);
                prob_z0 += log_prob_z0[feature_index];
                prob_z0 = numpy.exp(prob_z0);

                # compute the log likelihood when Znk=1
                self._Z[object_index, feature_index] = 1;
                prob_z1 = self.log_likelihood_X(self._X[[object_index], :], self._Z[[object_index], :]);
                prob_z1 += log_prob_z1[feature_index]
                prob_z1 = numpy.exp(prob_z1);

                Znk_is_0 = prob_z0 / (prob_z0 + prob_z1);
                if random.random() < Znk_is_0:
                    self._Z[object_index, feature_index] = 0;
                else:
                    self._Z[object_index, feature_index] = 1;

        return singleton_features;

    """
    sample K_new using metropolis hastings algorithm
    """
    def metropolis_hastings_K_new(self, object_index, singleton_features):
        if type(object_index) != list:
            object_index = [object_index];

        # sample K_new from the metropolis hastings proposal distribution, i.e., a poisson distribution with mean \frac{\alpha}{N}
        K_new = scipy.stats.poisson.rvs(self._alpha / self._N);
        K_old = len(singleton_features);

        if K_new <= 0 and K_old <= 0:
            return False;

        Z_i_tmp = self._Z[object_index, :];
        Z_i_tmp[:, singleton_features] = 0;
        X_residue = self._X[object_index, :] - numpy.dot(Z_i_tmp, self._A);

        log_new_old = 0;
        for d in xrange(self._D):
            log_new_old -= 0.5 * numpy.log((self._sigma_x ** 2 + K_new * self._sigma_a ** 2) / (self._sigma_x ** 2 + K_old * self._sigma_a ** 2));
            log_new_old -= 0.5 * X_residue[0, d] ** 2 * (1 / (self._sigma_x ** 2 + K_new * self._sigma_a ** 2) - 1 / (self._sigma_x ** 2 + K_old * self._sigma_a ** 2));

        accept_new = 1.0 / (1.0 + 1.0 / numpy.exp(log_new_old));

        '''
        # compute the log likelihood if we use old features
        Z_i_old = numpy.ones((1, K_old));
        M_old = self.compute_M(Z_i_old);
        assert(M_old.shape==(K_old, K_old));
        log_likelihood_old = 1-numpy.dot(numpy.dot(Z_i_old, M_old), Z_i_old.transpose());
        log_likelihood_old = -numpy.trace(numpy.dot(numpy.dot(X_residue.transpose(), log_likelihood_old), X_residue));
        log_likelihood_old /= (2 * self._sigma_x**2);
        log_likelihood_old += self._D / 2 * numpy.linalg.det(M_old);
        log_likelihood_old -= (1-K_old)*self._D * numpy.log(self._sigma_x) + (K_old*self._D) * numpy.log(self._sigma_a);

        # compute the log likelihood if we use new features
        Z_i_new = numpy.ones((1, K_new));
        M_new = self.compute_M(Z_i_new);
        assert(M_new.shape==(K_new, K_new));
        log_likelihood_new = 1-numpy.dot(numpy.dot(Z_i_new, M_new), Z_i_new.transpose());
        log_likelihood_new = -numpy.trace(numpy.dot(numpy.dot(X_residue.transpose(), log_likelihood_new), X_residue));
        log_likelihood_new /= (2 * self._sigma_x**2);
        log_likelihood_new += self._D / 2 * numpy.linalg.det(M_new);
        log_likelihood_new -= (1-K_new)*self._D * numpy.log(self._sigma_x) + (K_new*self._D) * numpy.log(self._sigma_a);

        # compute the probability of accepting new features
        accept_new = 1.0/(1.0 + numpy.exp(log_likelihood_old-log_likelihood_new));
        '''

        self._A = self._A[[k for k in xrange(self._K) if k not in singleton_features], :];
        self._Z = self._Z[:, [k for k in xrange(self._K) if k not in singleton_features]];
        self._K -= K_old

        # if we accept the proposal, we will replace old A and Z matrices
        if random.random() > accept_new:
            K_new = K_old;

        if K_new > 0:
            # construct A_new and Z_new
            Z_new = numpy.zeros((self._N, K_new));
            Z_i_new = numpy.ones((1, K_new));
            Z_new[object_index, :] = Z_i_new;
            M_new = self.compute_M(Z_i_new);
            A_new = numpy.dot(M_new, numpy.dot(Z_new.transpose(), self._X - numpy.dot(self._Z, self._A)));
            self._A = numpy.vstack((self._A, A_new));
            self._Z = numpy.hstack((self._Z, Z_new));
            #self._Z[object_index, :] = Z_i_new;
            self._K += K_new
            return True;

        return False;

    """
    """
    def sample_A(self):
        # sample every feature
        order = numpy.random.permutation(self._D);
        for (observation_counter, observation_index) in enumerate(order):
            # sample A_d
            (mean, std_dev) = self.sufficient_statistics_A([observation_index]);
            assert(std_dev.shape == (self._K, self._K));
            assert(mean.shape == (self._K, len([observation_index])));
            self._A[:, [observation_index]] = numpy.dot(std_dev, numpy.random.normal(0, 1, (self._K, len([observation_index])))) + mean;

        return

    """
    compute the mean and co-variance, i.e., sufficient statistics, of A
    @param observation_index: a list data type, recorded down the observation indices (column numbers) of A we want to compute
    """
    def sufficient_statistics_A(self, observation_index=None):
        if observation_index == None:
            X = self._X;
            observation_index = range(self._D);
        else:
            X = self._X[:, observation_index]

        assert(type(observation_index) == list);

        D = X.shape[1];
        #mean_a = numpy.zeros((self._K, D));
        #for k in range(self._K):
        #    mean_a[k, :] = self._mean_a[0, observation_index];
        A_prior = numpy.tile(self._A_prior[0, observation_index], (self._K, 1));

        assert(X.shape == (self._N, D));
        assert(self._Z.shape == (self._N, self._K));
        assert(A_prior.shape == (self._K, D))

        # compute M = (Z' * Z - (sigma_x^2) / (sigma_a^2) * I)^-1
        M = self.compute_M();
        # compute the mean of the matrix A
        mean_A = numpy.dot(M, numpy.dot(self._Z.transpose(), X) + (self._sigma_x / self._sigma_a) ** 2 * A_prior);
        # compute the co-variance of the matrix A
        std_dev_A = numpy.linalg.cholesky(self._sigma_x ** 2 * M).transpose();

        return (mean_A, std_dev_A)

    """
    remove the empty column in matrix Z and the corresponding feature in A
    """
    def regularize_matrices(self):
        assert(self._Z.shape == (self._N, self._K));
        Z_sum = numpy.sum(self._Z, axis=0);
        assert(len(Z_sum) == self._K);
        indices = numpy.nonzero(Z_sum == 0);
        #assert(numpy.min(indices)>=0 and numpy.max(indices)<self._K);

        #print self._K, indices, [k for k in range(self._K) if k not in indices]
        self._Z = self._Z[:, [k for k in range(self._K) if k not in indices]];
        self._A = self._A[[k for k in range(self._K) if k not in indices], :];

        self._K = self._Z.shape[1];
        assert(self._Z.shape == (self._N, self._K));
        assert(self._A.shape == (self._K, self._D));

    """
    compute the log-likelihood of the data X
    @param X: a 2-D numpy array
    @param Z: a 2-D numpy boolean array
    @param A: a 2-D numpy array, integrate A out if it is set to None
    """
    def log_likelihood_X(self, X=None, Z=None, A=None):
        if A == None:
            A = self._A;
        if Z == None:
            Z = self._Z;
        if X == None:
            X = self._X;

        assert(X.shape[0] == Z.shape[0]);
        (N, D) = X.shape;
        (N, K) = Z.shape;
        assert(A.shape == (K, D));

        log_likelihood = X - numpy.dot(Z, A);

        (row, column) = log_likelihood.shape;
        if row > column:
            log_likelihood = numpy.trace(numpy.dot(log_likelihood.transpose(), log_likelihood));
        else:
            log_likelihood = numpy.trace(numpy.dot(log_likelihood, log_likelihood.transpose()));

        log_likelihood = -0.5 * log_likelihood / numpy.power(self._sigma_x, 2);
        log_likelihood -= N * D * 0.5 * numpy.log(2 * numpy.pi * numpy.power(self._sigma_x, 2));

        return log_likelihood

    """
    compute the log-likelihood of A
    """
    def log_likelihood_A(self):
        log_likelihood = -0.5 * self._K * self._D * numpy.log(2 * numpy.pi * self._sigma_a * self._sigma_a);
        #for k in range(self._K):
        #    A_prior[k, :] = self._mean_a[0, :];
        A_prior = numpy.tile(self._A_prior, (self._K, 1))
        log_likelihood -= numpy.trace(numpy.dot((self._A - A_prior).transpose(), (self._A - A_prior))) * 0.5 / (self._sigma_a ** 2);

        return log_likelihood;

    """
    compute the log-likelihood of the model
    """
    def log_likelihood_model(self):
        #print self.log_likelihood_X(self._X, self._Z, self._A), self.log_likelihood_A(), self.log_likelihood_Z();
        return self.log_likelihood_X() + self.log_likelihood_A() + self.log_likelihood_Z();

    """
    sample noise variances, i.e., sigma_x
    """
    def sample_sigma_x(self, sigma_x_hyper_parameter):
        return self.sample_sigma(self._sigma_x_hyper_parameter, self._X - numpy.dot(self._Z, self._A));

    """
    sample feature variance, i.e., sigma_a
    """
    def sample_sigma_a(self, sigma_a_hyper_parameter):
        return self.sample_sigma(self._sigma_a_hyper_parameter, self._A - numpy.tile(self._A_prior, (self._K, 1)));

"""
run IBP on the synthetic 'cambridge bars' dataset, used in the original paper.
"""
if __name__ == '__main__':
    import scipy.io;
    #import util.scaled_image;

    # load the data from the matrix
    mat_vals = scipy.io.loadmat('../../data/cambridge-bars/block_image_set.mat');
    true_weights = mat_vals['trueWeights']
    features = mat_vals['features']
    data = mat_vals['data']

    print( true_weights.shape, features.shape, data.shape)

    # set up the hyper-parameter for sampling alpha
    alpha_hyper_parameter = (1., 1.);
    # set up the hyper-parameter for sampling sigma_x
    sigma_x_hyper_parameter = (1., 1.);
    # set up the hyper-parameter for sampling sigma_a
    sigma_a_hyper_parameter = (1., 1.);

    features = features.astype(numpy.int);

    # initialize the model
    ibp = SemicollapsedGibbsSampling(alpha_hyper_parameter, sigma_x_hyper_parameter, sigma_a_hyper_parameter, True);

    ibp._initialize(data[1:100, :], 0.5, 0.2, 0.5, None, None, None);

    ibp.sample(20);

    print( ibp._Z.sum(axis=0))

    # If matplotlib is installed, plot ground truth vs learned factors
    import matplotlib.pyplot as P
    from util.scaled_image import scaledimage

    # Intensity plots of
    # -ground truth factor-feature weights (top)
    # -learned factor-feature weights (bottom)
    K = max(len(true_weights), len(ibp._A))
    (fig, subaxes) = P.subplots(2, K)
    for sa in subaxes.flatten():
        sa.set_visible(False)
    fig.suptitle('Ground truth (top) vs learned factors (bottom)')
    for (idx, trueFactor) in enumerate(true_weights):
        ax = subaxes[0, idx]
        ax.set_visible(True)
        scaledimage(trueFactor.reshape(6, 6),
                    pixwidth=3, ax=ax)
    for (idx, learnedFactor) in enumerate(ibp._A):
        ax = subaxes[1, idx]
        scaledimage(learnedFactor.reshape(6, 6),
                    pixwidth=3, ax=ax)
        ax.set_visible(True)
    P.show()



# -*- coding: utf-8 -*-

from pymake import ExpDesign, ExpSpace, ExpTensor, Corpus, Model, ExpVector

__plot_font_size = 14

### PLOTLIB CONFIG
import matplotlib.pyplot as plt
plt.rc('font', size=__plot_font_size)  # controls default text sizes

class Netw(ExpDesign):

    # Use for Name on figure and table
    _alias = dict((
        ('propro'   , 'Protein')  ,
        ('blogs'    , 'Blogs')    ,
        ('euroroad' , 'Euroroad') ,
        ('emaileu'  , 'Emaileu') ,
        ('manufacturing'  , 'Manufacturing'),
        ('fb_uc'          , 'UC Irvine' ),
        ('generator7'     , 'Network1' ),
        ('generator12'    , 'Network2' ),
        ('generator10'    , 'Network3' ),

        #('generator4'     , 'Network4' ),
        ('generator4'     , 'Network2' ),
        ('ml.ilfm_cgs'     , 'ILFM' ),
        ('ml.immsb_cgs'     , 'IMMSB' ),
    ))


    # Networks Data
    corpus_real_net = Corpus(['manufacturing', 'fb_uc','blogs', 'emaileu', 'propro', 'euroroad'])

    ### Bursty
    CORPUS_BURST_1     = Corpus(['generator3', 'generator11', 'generator12', 'generator7', 'generator14'])

    ### Non Bursty
    CORPUS_NBURST_1    = Corpus(['generator4', 'generator5', 'generator6', 'generator9', 'generator10'])

    CORPUS_SYN_ICDM  = Corpus(['generator7', 'generator12', 'generator10', 'generator4'])
    CORPUS_REAL_ICDM = Corpus(['manufacturing', 'fb_uc',])
    CORPUS_ALL_ICDM = CORPUS_SYN_ICDM + CORPUS_REAL_ICDM
    CORPUS_REAL_PNAS = Corpus(['manufacturing', 'fb_uc','blogs', 'emaileu', 'propro'])
    CORPUS_ALL_PNAS = CORPUS_REAL_PNAS +  CORPUS_SYN_ICDM
    pnas_short = Corpus([ 'blogs', 'manufacturing', 'generator7','generator4'])
    pnas_rest = (corpus_real_net + CORPUS_SYN_ICDM) - pnas_short

    #data_net_all
    #net_all

    # Text Corpus
    # intruder ?
    data_text_all = Corpus(['kos', 'nips12', 'nips', 'reuter50', '20ngroups']) # lucene

    # Tensor Exp
    EXPE_ICDM = ExpTensor((
        ('_data_type', ('networks',)),
        ('_refdir'  , ('debug10', 'debug11')),
        #('corpus' , ('fb_uc', 'manufacturing')),
        ('corpus' , CORPUS_ALL_ICDM),
        ('model'  , ('immsb_cgs', 'ilfm_cgs')),
        ('K'      , (5,10,15,20)),
        ('N'      , ('all',)),
        ('hyper'  , ('fix', 'auto')),
        ('homo'   , (0,)),
        #('_repeat'   , (0, 1, 2,3, 4, 5)),
    ))

    PNAS1 = ExpTensor ((
        ('corpus', CORPUS_ALL_PNAS),
        ('_data_type'    , 'networks'),
        ('_refdir'        , 'debug111111') , # ign in gen
        #('model'        , 'mmsb_cgs')   ,
        ('model'        , ['immsb_cgs', 'ilfm_cgs'])   ,
        ('K'            , 10)        ,
        ('N'            , 'all')     , # ign in gen
        ('hyper'        , ['auto', 'fix'])    , # ign in gen
        ('homo'         , 0)         , # ign in gen
        ('_repeat'      , 1)       ,
        ('_bind'    , ['immsb_cgs.auto', 'ilfm_cgs.fix']),
        ('iterations', '200'),
        ('_format', '{model}_{corpus}_{K}_{hyper}_{homo}_{N}')
    ))

    PNAS2 = ExpTensor ((
        ('corpus', CORPUS_ALL_PNAS),
        ('_data_type'    , 'networks'),
        ('_refdir'        , 'pnas2') , # ign in gen
        #('model'        , 'mmsb_cgs')   ,
        ('model'        , ['immsb_cgs', 'ilfm_cgs'])   ,
        ('K'            , 10)        ,
        ('N'            , 'all')     , # ign in gen
        ('hyper'        , ['fix','auto'])    , # ign in gen
        ('homo'         , 0)         , # ign in gen
        ('_repeat'      , 0)       ,
        ('_bind'    , ['immsb_cgs.auto', 'ilfm_cgs.fix', 'ilfm_cgs.iterations.25', 'immsb_cgs.iterations.150']),
        ('iterations', [25, 150]),
        ('testset_ratio', [40, 60, 80]),
        ('_format', '{model}_{corpus}_{K}_{hyper}_{homo}_{N}_{testset_ratio}'),
        ('_csv_typo', '_iteration time_it _entropy _entropy_t _K _alpha _gmma alpha_mean delta_mean alpha_var delta_var'),

    ))

    PNAS3 = ExpTensor ((
        ('corpus', pnas_short),
        ('_refdir'        , 'pnas3') , # ign in gen
        ('_data_type'    , 'networks'),
        ('model'        , ['immsb_cgs', 'ilfm_cgs'])   ,
        ('K'            , 10)        ,
        ('N'            , 'all')     , # ign in gen
        ('hyper'        , ['fix','auto'])    , # ign in gen
        ('homo'         , 0)         , # ign in gen
        ('_repeat'      , 0)       ,
        ('_bind'    , ['immsb_cgs.auto', 'ilfm_cgs.fix', 'ilfm_cgs.iterations.25', 'immsb_cgs.iterations.150']),
        ('iterations', [25, 150]),
        ('testset_ratio', 20),
        ('_format', '{model}_{iterations}_{corpus}_{K}_{hyper}_{homo}_{N}_{testset_ratio}'),
        #('_csv_typo', '_iteration time_it _entropy _entropy_t _K _alpha _gmma alpha_mean delta_mean alpha_var delta_var'),

    ))

    EXPE_ICDM_R = ExpTensor((
        ('_data_type', ('networks',)),
        #('corpus' , ('fb_uc', 'manufacturing')),
        ('corpus' , CORPUS_SYN_ICDM),
        #('_refdir'  , ('debug10', 'debug11')),
        ('_refdir'  , ('debug101010', 'debug111111')),
        ('model'  , ('immsb_cgs', 'ilfm_cgs')),
        ('K'      , (5, 10, 15, 20)),
        ('hyper'  , ('fix', 'auto')),
        ('homo'   , (0, 1, 2)),
        ('N'      , ('all',)),
        ('_repeat'   , list(range(10))),
        ('iterations', '200'),
        ('_bind'    , ['immsb_cgs.auto', 'ilfm_cgs.fix']),
    ))

    EXPE_ICDM_R_R = ExpTensor((
        ('_data_type', ('networks',)),
        ('corpus' , ('fb_uc', 'manufacturing')),
        ('_refdir'  , ('debug101010', 'debug111111')),
        ('model'  , ('immsb_cgs', 'ilfm_cgs')),
        ('K'      , (5, 10, 15, 20)),
        ('hyper'  , ('fix', 'auto')),
        ('homo'   , (0, 1, 2)),
        ('N'      , ('all',)),
        ('_repeat'   , list(range(10))),
        ('iterations', '200'),
        ('_bind'    , ['immsb_cgs.auto', 'ilfm_cgs.fix']),
    ))

    # Single Expe

    MODEL_FOR_CLUSTER_IBP = ExpSpace((
        ('_data_type'    , 'networks'),
        ('_refdir'        , 'debug11') ,
        ('model'        , 'ilfm_cgs')   ,
        ('K'            , 20)        ,
        ('N'            , 'all')     ,
        ('hyper'        , 'fix')     ,
        ('homo'         , 0)         ,
        #('_repeat'      , '*')       ,
    ))
    MODEL_FOR_CLUSTER_IMMSB = ExpSpace((
        ('_data_type'    , 'networks'),
        ('_refdir'        , 'debug11') ,
        ('model'        , 'immsb_cgs')   ,
        ('K'            , 20)        ,
        ('N'            , 'all')     ,
        ('hyper'        , 'auto')     ,
        ('homo'         , 0)         ,
        #('_repeat'      , '*')       ,
    ))

    default_gen = ExpTensor ((
        ('corpus', CORPUS_SYN_ICDM),
        ('_data_type'    , 'networks'),
        ('_refdir'        , 'debug111111') , # ign in gen
        #('model'        , 'mmsb_cgs')   ,
        ('model'        , ['immsb_cgs', 'ilfm_cgs'])   ,
        ('K'            , 10)        ,
        ('N'            , 'all')     , # ign in gen
        ('hyper'        , ['auto', 'fix'])    , # ign in gen
        ('homo'         , 0)         , # ign in gen
        ('_repeat'      , 1)       ,
        ('_bind'    , ['immsb_cgs.auto', 'ilfm_cgs.fix']),
        ('alpha', 1),
        ('gmma', 1),
        ('delta', [(1, 5)]),
    ))
    default_check = default_gen.copy()
    default_check['model'] = 'immsb_cgs'


    MODELS_GENERATE = ExpTensor ((
        ('_data_type'    , 'networks'),
        ('_refdir'        , 'debug11') ,
        ('model'        , ['immsb_cgs', 'ilfm_cgs'])   ,
        ('K'            , 10)        ,
        ('N'            , 'all')     ,
        ('hyper'        , ['fix', 'auto'])     ,
        ('homo'         , 0)         ,
        #('_repeat'      , '*')       ,
        ('_bind'    , ['immsb_cgs.auto', 'ilfm_cgs.fix']),
        ('_csv_typo ', ('_iteration time_it _entropy _entropy_t _K _alpha _gmma alpha_mean delta_mean alpha_var delta_var',)),
    ))


#### Temp

    EXPE_ALL_ICDM_IBP = ExpTensor((
        ('_data_type', ('networks',)),
        ('_refdir'  , ('debug111111', 'debug101010')),
        ('corpus' , CORPUS_ALL_ICDM),
        ('model'  , ('ilfm_cgs',)),
        ('K'      , (5, 10, 15, 20)),
        ('N'      , ('all',)),
        ('hyper'  , ('fix',)),
        ('homo'   , (0,)),
        ('_repeat'   , (6, 7, 8, 9)),
        ('_csv_typo ', ('_iteration time_it _entropy _entropy_t _K _alpha _gmma alpha_mean delta_mean alpha_var delta_var',)),

    ))
    EXPE_ALL_ICDM_IMMSB = ExpTensor((
        ('_data_type', ('networks',)),
        ('_refdir'  , ('debug111111', 'debug101010')),
        ('corpus' , CORPUS_ALL_ICDM),
        ('model'  , ('immsb_cgs',)),
        ('K'      , (5, 10, 15, 20)),
        ('N'      , ('all',)),
        ('hyper'  , ('auto',)),
        ('homo'   , (0,)),
        ('_repeat'   , (6, 7, 8, 9)),
        ('_csv_typo ', ('_iteration time_it _entropy _entropy_t _K _alpha _gmma alpha_mean delta_mean alpha_var delta_var',)),
    ))



    RUN_DD = ExpTensor((
        ('_data_type', ('networks',)),
        #('corpus' , ('fb_uc', 'manufacturing')),
        ('_refdir' , ('test_temp',)),
        ('corpus' , ('generator1',)),
        ('model'  , ('immsb_cgs', 'ilfm_cgs')),
        ('K'      , (5,)),
        ('N'      , ('all',)),
        ('hyper'  , ('auto',)),
        ('homo'   , (0,)),
        ('hyper_prior', ('1 2 3 4', '20 2 10 2')),
        ('_repeat'   , (0, 1, 2, 4, 5)),
        ('_bind'    , ['immsb_cgs.auto', 'ilfm_cgs.fix']),
        ('_csv_typo ', ('_iteration time_it _entropy _entropy_t _K _alpha _gmma alpha_mean delta_mean alpha_var delta_var',)),
    ))

    EXPE_REAL_V2_IBP = ExpTensor((
        ('_data_type', ('networks',)),
        ('corpus' , ( 'propro', 'blogs', 'euroroad', 'emaileu')),
        ('_refdir'  , ('debug111111'),),
        ('model'  , ( 'ilfm_cgs',)),
        ('K'      , ( 10,)),
        ('hyper'  , ('fix',)),
        ('homo'   , (0,)),
        ('N'      , ('all',)),
        ('_repeat'   , list(range(5))),
        ('_csv_typo ', ('_iteration time_it _entropy _entropy_t _K _alpha _gmma alpha_mean delta_mean alpha_var delta_var',)),
    ))

    EXPE_REAL_V2_IMMSB = ExpTensor((
        ('_data_type', ('networks',)),
        ('corpus' , ( 'propro', 'blogs', 'euroroad', 'emaileu')),
        ('_refdir'  , ('debug111111',),),
        ('model'  , ( 'immsb_cgs',)),
        ('K'      , ( 10,)),
        ('hyper'  , ('auto',)),
        ('homo'   , (0,)),
        ('N'      , ('all',)),
        ('_repeat'   , list(range(5))),
        ('_csv_typo ', ('_iteration time_it _entropy _entropy_t _K _alpha _gmma alpha_mean delta_mean alpha_var delta_var',)),
    ))

    RAGNRK = ExpTensor(
        _data_type = ['networks'],
        corpus = ['propro', 'blogs', 'euroroad', 'emaileu'],
        _refdir  = ['ragnarok'],
        model  = ['immsb_cgs'],
        K      = [10],
        hyper  = ['auto'],
        homo   = [0],
        N      = [10],
        _repeat = list(range(2)),
        _csv_typo = '_iteration time_it _entropy _entropy_t _K _alpha _gmma alpha_mean delta_mean alpha_var delta_var'
    )


    default_expe = ExpTensor(
        _data_type   = 'networks',
        corpus      = 'clique4',
        model       = 'immsb_cgs',
        hyper       = 'auto',
        _refdir      = 'debug',
        testset_ratio = 20,
        K           = 5,
        N           = 50,
        iterations  = 10,
        homo = 0,
        _format = '{model}_{corpus}_{K}_{hyper}_{homo}_{N}',
        #_csv_typo = '_iteration time_it _entropy _entropy_t _K _alpha _gmma alpha_mean delta_mean alpha_var delta_var'
    )

    debug = ExpTensor(
        _data_type   = 'networks',
        corpus      = 'clique4',
        model       = ['immsb_cgs', 'ilfm_cgs', 'immsb_scvb'],
        hyper       = 'auto',
        _refdir      = 'debug',
        testset_ratio = 20,
        K           = 4,
        N           = 42,
        iterations  = 3,
        homo = 0,
        _format = '{model}_{corpus}_{K}_{hyper}_{homo}_{N}',
        _csv_typo = '_iteration time_it _entropy _entropy_t _K'
    )



    model_all =  Model(['immsb_cgs', 'ilfm_cgs', 'immsb_scvb', 'rescal'])
    model_debug =  Model(['mmsb_vb'])



